import os
import re
import tempfile

from scm.plams.interfaces.molecule.ase import toASE
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.kftools import KFFile
from scm.plams.tools.units import Units
from scm.plams.trajectories.rkffile import RKFTrajectoryFile
from scm.plams.trajectories.rkfhistoryfile import RKFHistoryFile

__all__ = [
    "traj_to_rkf",
    "vasp_output_to_ams",
    "qe_output_to_ams",
    "gaussian_output_to_ams",
    "rkf_to_ase_traj",
    "rkf_to_ase_atoms",
    "file_to_traj",
]


def traj_to_rkf(trajfile, rkftrajectoryfile, task=None, timestep: float = 0.25):
    """
    Convert ase .traj file to .rkf file. NOTE: The order of atoms (or the number of atoms) cannot change between frames!

    trajfile : str
        path to a .traj file
    rkftrajectoryfile : str
        path to the output .rkf file (will be created)
    task : str
        Which task to write. If None it is auto-determined.

    timestep: float
        Which timestep to write when task == 'moleculardynamics'


    Returns : 2-tuple (coords, cell)
        The final coordinates and cell in angstrom
    """
    import warnings

    from ase.io import Trajectory

    warnings.filterwarnings("ignore", "Creating an ndarray from ragged nested sequences")
    traj = Trajectory(trajfile)
    rkfout = RKFTrajectoryFile(rkftrajectoryfile, mode="wb")
    rkfout.store_historydata()
    rkfout.store_mddata()

    # Store the units properly
    mdunits = {}
    mdunits["PotentialEnergy"] = "hartree"
    mdunits["KineticEnergy"] = "hartree"
    mdunits["TotalEnergy"] = "hartree"
    rkfout._set_mdunits(mdunits)

    energy_converter = Units.convert(1.0, "eV", "hartree")
    gradients_converter = Units.convert(1.0, "eV/angstrom", "hartree/bohr")
    stress_converter = Units.convert(1.0, "eV/angstrom^3", "hartree/bohr^3")

    coords, cell = None, None
    try:
        for i, atoms in enumerate(traj):
            if i == 0:
                rkfout.set_elements(atoms.get_chemical_symbols())
            coords = atoms.get_positions()  # angstrom
            cell = atoms.get_cell()
            try:
                gradients = -atoms.get_forces() * gradients_converter
            except:
                gradients = None
            try:
                stresstensor = atoms.get_stress(voigt=False) * stress_converter
            except:
                stresstensor = None
            mddata = {}
            try:
                energy = atoms.get_potential_energy() * energy_converter
                mddata["PotentialEnergy"] = energy
            except:
                energy = None

            try:
                kinetic_energy = atoms.get_kinetic_energy() * energy_converter
                mddata["KineticEnergy"] = kinetic_energy
            except:
                kinetic_energy = None

            if "PotentialEnergy" in mddata and "KineticEnergy" in mddata:
                mddata["TotalEnergy"] = mddata["PotentialEnergy"] + mddata["KineticEnergy"]

            if str(task).lower() == "moleculardynamics" or len(mddata) > 0:
                mddata["Time"] = timestep * i

            if len(mddata) == 0:
                mddata = None

            # Create a historydata dictionary, to go into the History section
            historydata = {}
            if gradients is not None:
                historydata["Gradients"] = gradients
                historydata["EngineGradients"] = gradients
            if stresstensor is not None:
                historydata["StressTensor"] = stresstensor
            if len(historydata) == 0:
                historydata = {}

            rkfout.write_next(coords=coords, cell=cell, historydata=historydata, mddata=mddata)

    finally:
        rkfout.close()

    # the below is needed to be able to load the .rkf file with AMSJob.load_external()
    kf = KFFile(rkftrajectoryfile)
    kf["EngineResults%nEntries"] = 0
    kf["General%program"] = "ams"
    if task is None:
        if len(traj) == 1:
            task = "singlepoint"
        elif kinetic_energy is None or kinetic_energy == 0:
            task = "geometryoptimization"
        else:
            task = "moleculardynamics"
    kf["General%task"] = task
    kf["General%user input"] = "\xFF".join([f"Task {task}", "Engine External", "EndEngine"])

    return coords, cell


def file_to_traj(outfile, trajfile):
    """
    outfile : str
        path to existing file (OUTCAR, qe.out, etc.)
    trajfile : str
        will be created
    """
    from ase.io import read, write

    if os.path.exists(trajfile):
        os.remove(trajfile)

    atoms = read(outfile, ":")
    write(trajfile, atoms)

    if not os.path.exists(trajfile):
        raise RuntimeError("Couldn't write {}".format(trajfile))

    return trajfile


def _remove_or_raise(file, overwrite):
    if os.path.exists(file):
        if overwrite:
            os.remove(file)
        else:
            raise RuntimeError("{} already exists, specify overwrite=True to overwrite".format(file))


def _write_engine_rkf(kffile, enginefile):
    kf = KFFile(kffile)
    enginerkf = KFFile(enginefile)
    # write engine.rkf
    # copy General, Molecule, InputMolecule from ams.rkf
    for sec in ["General", "Molecule", "InputMolecule"]:
        secdict = kf.read_section(sec)
        for k, v in secdict.items():
            enginerkf[sec + "%" + k] = v
    enginerkf["General%program"] = "plams"
    nEntries = kf["History%nEntries"]
    suffix = "({})".format(nEntries)
    if ("History", "Energy" + suffix) in kf:
        enginerkf["AMSResults%Energy"] = kf["History%Energy" + suffix]
    if ("History", "Gradients" + suffix) in kf:
        enginerkf["AMSResults%Gradients"] = kf["History%Gradients" + suffix]
    if ("History", "StressTensor" + suffix) in kf:
        enginerkf["AMSResults%StressTensor"] = kf["History%StressTensor" + suffix]


def _postprocess_vasp_amsrkf(kffile, outcar):
    # add extra info to the kffile
    kf = KFFile(kffile, autosave=False)
    try:
        kf["EngineResults%nEntries"] = 1
        kf["EngineResults%Title(1)"] = "vasp"
        kf["EngineResults%Description(1)"] = "Standalone VASP run. Data from {}".format(os.path.abspath(outcar))
        kf["EngineResults%Files(1)"] = "vasp.rkf"
        kf["General%user input"] = "!VASP"

        # read the INCAR
        incarfile = os.path.join(os.path.dirname(outcar), "INCAR")
        userinput = ["!VASP", "Engine External", "  Input", "  !INCAR"]
        if os.path.exists(incarfile):
            with open(incarfile) as incar:
                for line in incar:
                    line = line.split("!")[0]
                    line = line.split("#")[0]
                    line = line.strip()
                    if line.lower().startswith("end"):  # "End" is reserved to end the block
                        line = "!" + line
                    if len(line) > 0:
                        userinput.append("    " + line)
            userinput.append("  !EndINCAR")
        userinput.append("  EndInput")  # end of the Free block
        userinput.append("EndEngine")
        userinput.append("Task {}".format(kf["General%task"]))
        kf["General%user input"] = "\xFF".join(userinput)

    finally:
        kf.save()


def vasp_output_to_ams(
    vasp_folder, wdir=None, overwrite=False, write_engine_rkf=True, task: str = None, timestep: float = 0.25
):
    """
    Converts VASP output (OUTCAR, ...) to AMS output (ams.rkf, vasp.rkf)

    Returns: a string containing the directory where ams.rkf was written

    vasp_folder : str
        path to a directory with an OUTCAR, INCAR, POTCAR etc. files

    wdir : str or None
        directory in which to write the ams.rkf and vasp.rkf files
        If None, a subdirectory "AMSJob" of vasp_folder will be created

    overwrite : bool
        if False, first check if wdir already contains ams.rkf and vasp.rkf, in which case do nothing
        if True, overwrite if exists

    write_engine_rkf : bool
        If True, also write vasp.rkf alongside ams.rkf. The vasp.rkf file will only contain an AMSResults section (energy, gradients, stress tensor). It will not contain the DOS or the band structure.

    task : str
        Which task to write to ams.rkf. If None it is auto-determined (probably set to 'geometryoptimization')

    timestep : float
        If task='moleculardynamics', which timestep (in fs) between frames to write
    """
    if not os.path.isdir(vasp_folder):
        raise ValueError("Directory {} does not exist".format(vasp_folder))

    outcar = os.path.join(vasp_folder, "OUTCAR")
    if not os.path.exists(outcar):
        if os.path.exists(os.path.join(vasp_folder, "XDATCAR")):
            outcar = os.path.join(vasp_folder, "XDATCAR")
        else:
            raise ValueError("File {} does not exist, should be an OUTCAR file.".format(outcar))

    if wdir is None:
        wdir = os.path.join(os.path.dirname(outcar), "AMSJob")
        os.makedirs(wdir, exist_ok=True)

    # exit early if ams.rkf already exists
    if os.path.exists(os.path.join(wdir, "ams.rkf")) and not overwrite:
        return wdir

    # convert OUTCAR to a .traj file inside wdir
    trajfile = file_to_traj(outcar, os.path.join(wdir, "vasp.traj"))

    # remove the target files first if overwrite
    kffile = os.path.join(wdir, "ams.rkf")
    enginefile = os.path.join(wdir, "vasp.rkf")
    _remove_or_raise(kffile, overwrite)
    _remove_or_raise(enginefile, overwrite)

    # convert the .traj file to ams.rkf
    traj_to_rkf(trajfile, kffile, task=task, timestep=timestep)

    _postprocess_vasp_amsrkf(kffile, outcar)
    if write_engine_rkf:
        _write_engine_rkf(kffile, enginefile)

    if os.path.exists(trajfile):
        os.remove(trajfile)

    return wdir


def _postprocess_qe_amsrkf(kffile, qe_outfile):
    # add extra info to the kffile
    kf = KFFile(kffile, autosave=False)
    try:
        kf["EngineResults%nEntries"] = 1
        kf["EngineResults%Title(1)"] = "qe"
        kf["EngineResults%Description(1)"] = "Standalone Quantum ESPRESSO run. Data from {}".format(
            os.path.abspath(qe_outfile)
        )
        kf["EngineResults%Files(1)"] = "qe.rkf"

        userinput = [
            "!QuantumESPRESSO",
            "Engine External",
            "  Input",
            "    Unknown Quantum ESPRESSO input",
            "  EndInput",
            "EndEngine",
        ]
        kf["General%user input"] = "\xFF".join(userinput)

    finally:
        kf.save()


def _postprocess_gaussian_amsrkf(kffile, gaussian_outfile):
    # add extra info to the kffile
    kf = KFFile(kffile, autosave=False)
    try:
        kf["EngineResults%nEntries"] = 1
        kf["EngineResults%Title(1)"] = "gaussian"
        kf["EngineResults%Description(1)"] = "Standalone Gaussian. Data from {}".format(
            os.path.abspath(gaussian_outfile)
        )
        kf["EngineResults%Files(1)"] = "gaussian.rkf"

        userinput = ["!Gaussian", "Engine External", "  Input", "    Unknown Gaussian input", "  EndInput", "EndEngine"]
        kf["General%user input"] = "\xFF".join(userinput)

    finally:
        kf.save()


def text_out_file_to_ams(qe_outfile, wdir=None, overwrite=False, write_engine_rkf=True, enginename="qe"):
    """
    Converts a qe .out or gaussian .out file to ams.rkf and qe.rkf/gaussian.rkf

    Do not use this function directly, instaead call qe_output_to_ams or gaussian_output_to_ams
    """
    if not os.path.exists(qe_outfile) or os.path.isdir(qe_outfile):
        raise FileNotFoundError(qe_outfile)

    basename = os.path.basename(qe_outfile)
    basename_no_suffix = basename
    if basename.endswith(".out"):
        basename_no_suffix = re.sub(".out$", "", basename)
    dirname = os.path.abspath(os.path.dirname(qe_outfile))

    if wdir is None:
        if os.path.isdir(os.path.join(dirname, basename_no_suffix + ".results")):
            wdir = os.path.join(dirname, basename_no_suffix + ".results", "AMSJob")
        else:
            wdir = os.path.join(dirname, basename_no_suffix + ".AMSJob")

    if os.path.exists(os.path.join(wdir, "ams.rkf")) and not overwrite:
        return wdir

    os.makedirs(wdir, exist_ok=True)

    # convert to a .traj file inside wdir
    # first trim the qe_outfile to the first occurrence of "JOB DONE". This is needed because
    # running standalone QE via the AMS GUI will print multiple jobs into the same output file
    # i.e. both the geo opt and band structure calculation into the same file, which causes
    # the ASE qe.out parser to crash
    with tempfile.NamedTemporaryFile(delete=False) as tf:
        with open(qe_outfile, "r") as instream:
            for line in instream:
                tf.write(line.encode())
                if "JOB DONE" in line:
                    break
        tf.flush()
        tf.file.close()
        trajfile = file_to_traj(tf.name, os.path.join(wdir, "out.traj"))
        os.remove(tf.name)

    # remove the target files first if overwrite
    kffile = os.path.join(wdir, "ams.rkf")
    enginefile = os.path.join(wdir, f"{enginename}.rkf")
    _remove_or_raise(kffile, overwrite)
    _remove_or_raise(enginefile, overwrite)

    # convert the .traj file to ams.rkf
    traj_to_rkf(trajfile, kffile)

    if write_engine_rkf:
        # here one could also run $AMSBIN/tokf to get things like the DOS
        # $AMSBIN/tokf qe qe_outfile enginefile
        # But that would need to reread the entire trajectory
        # and one would need to postprocess it with the AMSResults section
        _write_engine_rkf(kffile, enginefile)

    if os.path.exists(trajfile):
        os.remove(trajfile)

    return wdir


def qe_output_to_ams(qe_outfile, wdir=None, overwrite=False, write_engine_rkf=True):
    """
    Converts a qe .out file to ams.rkf and qe.rkf.

    Returns: a string containing the directory where ams.rkf was written

    If the filename ends in .out, check if a .results directory exists. In that case, place
    the AMSJob subdirectory in the .results directory.

    Otherwise, create a new directory called filename.AMSJob

    qe_outfile : str
        path to the qe output file

    """
    wdir = text_out_file_to_ams(
        qe_outfile, wdir, overwrite=overwrite, write_engine_rkf=write_engine_rkf, enginename="qe"
    )
    _postprocess_qe_amsrkf(os.path.join(wdir, "ams.rkf"), qe_outfile)
    return wdir


def gaussian_output_to_ams(outfile, wdir=None, overwrite=False, write_engine_rkf=True):
    """
    Converts a Gaussian .out file to ams.rkf and gaussian.rkf.

    Returns: a string containing the directory where ams.rkf was written

    If the filename ends in .out, check if a .results directory exists. In that case, place
    the AMSJob subdirectory in the .results directory.

    Otherwise, create a new directory called filename.AMSJob

    outfile : str
        path to the gaussian output file

    """
    wdir = text_out_file_to_ams(
        outfile, wdir, overwrite=overwrite, write_engine_rkf=write_engine_rkf, enginename="gaussian"
    )
    _postprocess_gaussian_amsrkf(os.path.join(wdir, "ams.rkf"), outfile)
    return wdir


def rkf_to_ase_atoms(rkf_file, get_results=True):
    """
    Convert an ams.rkf trajectory to a list of ASE atoms

    rkf_file: str
        Path to an ams.rkf file

    get_results: bool
        Whether to include results like energy, forces, and stress in the trajectory.

    Returns: a list of all the ASE Atoms objects.
    """
    from ase import Atoms
    from ase.calculators.singlepoint import SinglePointCalculator
    import numpy as np

    bohr2angstrom = Units.convert(1.0, "bohr", "angstrom")
    hartree2eV = Units.convert(1.0, "hartree", "eV")

    def get_ase_atoms(elements, crd, cell, energy, gradients, stress):

        pbc = None
        if cell is not None:
            cell = np.array(cell).reshape(-1, 3)
            pbc = ["T"] * len(cell) + ["F"] * (3 - len(cell))
        atoms = Atoms(symbols=elements, positions=np.array(crd).reshape(-1, 3), cell=cell, pbc=pbc)
        if get_results:
            calculator = SinglePointCalculator(atoms)
            atoms.set_calculator(calculator)

            if energy:
                atoms.calc.results["energy"] = energy * hartree2eV
            if gradients:
                forces = -np.array(gradients).reshape(-1, 3) * hartree2eV / bohr2angstrom
                atoms.calc.results["forces"] = forces
            if stress:
                n = len(stress)
                if n == 9:
                    stress = np.array(stress).reshape(3, 3) * hartree2eV / bohr2angstrom**3
                    atoms.calc.results["stress"] = np.array(
                        [stress[0][0], stress[1][1], stress[2][2], stress[1][2], stress[0][2], stress[0][1]]
                    )

        return atoms

    rkf_filename = rkf_file
    kf = KFFile(rkf_filename)
    if "History" in kf.keys():
        if "ChemicalSystem(1)" in kf.keys():
            rkf = RKFHistoryFile(rkf_filename)
        else:
            rkf = RKFTrajectoryFile(rkf_filename)

        rkf.store_historydata()
        all_atoms = []
        for crd, cell in rkf:
            energy, stress = None, None
            if get_results:
                energy = rkf.historydata.get("EngineEnergy", None)
                if energy is None:
                    energy = rkf.historydata.get("Energy", None)
                gradients = rkf.historydata.get("EngineGradients", None)
                if gradients is None:
                    gradients = rkf.historydata.get("Gradients", None)
                stress = rkf.historydata.get("StressTensor", None)
            atoms = get_ase_atoms(rkf.elements, crd, cell, energy, gradients, stress)
            all_atoms.append(atoms)
    else:
        atoms = toASE(Molecule(rkf_filename))
        all_atoms = [atoms]

    return all_atoms


def rkf_to_ase_traj(rkf_file, out_file, get_results=True):
    """
    Convert an ams.rkf trajectory to a different trajectory format (.xyz, .traj, anything supported by ASE)

    rkf_file: str
        Path to an ams.rkf file

    out_file: str
        Path to the .traj or .xyz file that will be created. If the file exists it will be overwritten. If a .xyz file is specified it will use the normal ASE format (not the AMS format).

    get_results: bool
        Whether to include results like energy, forces, and stress in the trajectory.

    """

    from ase.io import write

    all_atoms = rkf_to_ase_atoms(rkf_file, get_results=get_results)
    write(out_file, all_atoms)
    return all_atoms



from scm.plams.core.errors import ResultsError
from scm.plams.core.functions import log
from scm.plams.interfaces.adfsuite.scmjob import SCMJob, SCMResults
from scm.plams.tools.units import Units

__all__ = ["DFTBJob", "DFTBResults"]


class DFTBResults(SCMResults):
    _kfext = ".rkf"
    _rename_map = {"dftb.rkf": "$JN" + _kfext}

    def get_main_molecule(self):
        """get_main_molecule()
        Return a |Molecule| instance based on the ``Molecule`` section in the main KF file (``$JN.rkf``).

        For runs with multiple geometries (geometry optimization, transition state search, molecular dynamics) this is the **final** geometry. In such a case, to access the initial (or any intermediate) coordinates please extract them from section ``History``, variable ``Coords(i)`` or use :meth:`get_input_molecule`. Mind the fact that all coordinates written by DFTB to ``.rkf`` file are in bohr::

            mol = results.get_molecule(section='History', variable='Coords(1)', unit='bohr')
        """
        ret = self.get_molecule(section="Molecule", variable="Coords", unit="bohr")
        ret.properties.charge = self.readkf("Molecule", "Charge")
        if ("Molecule", "LatticeVectors") in self._kf:
            lattice = self.readkf("Molecule", "LatticeVectors")
            lattice = Units.convert(lattice, "bohr", "angstrom")
            ret.lattice = [tuple(lattice[i : i + 3]) for i in range(0, len(lattice), 3)]
        return ret

    def get_input_molecule(self):
        """get_input_molecule()
        Return a |Molecule| instance with initial coordinates.

        All data used by this method is taken from ``$JN.rkf`` file. The ``molecule`` attribute of the corresponding job is ignored.
        """
        if ("History", "nEntries") in self._kf:
            return self.get_molecule(section="History", variable="Coords(1)", unit="bohr")
        return self.get_main_molecule()

    def get_history_molecules(self):
        """get_history_molecules()
        Return a list of |Molecule| instances with all Structures from the ``History`` section of the ``.rkf`` file.
        If the structures have lattice information, it is written into the corresponding |Molecule| instance.

        All data used by this method is taken from ``$JN.rkf`` file. If no ``History`` section is available an empty list is returned and a level 5 log entry appears.
        """
        if not ("History", "nEntries") in self._kf:
            log("No History section found in rkf during get_history_molecules, returning an empty list.", level=5)
            return []

        nEntries = self.readkf("History", "nEntries")
        # list of molecules to return
        history = []
        # iterate over entries
        for i in range(1, nEntries + 1):
            mol = self.get_molecule("History", "Coords(" + str(i) + ")")
            # read lattice if there is any
            if ("History", "LatticeVectors(" + str(i) + ")") in self._kf:
                # LatticeVectors has to be length dimensions*3, so we can do this
                lattice = Units.convert(self.readkf("History", "LatticeVectors(" + str(i) + ")"), "bohr", "angstrom")
                mol.lattice = [tuple(lattice[j : j + 3]) for j in range(0, len(lattice), 3)]
            history.append(mol)
        return history

    def get_energy(self, unit="au"):
        """get_energy(unit='au')
        Return DFTB final energy, expressed in *unit*.
        """
        prop = self.get_properties()
        if "DFTB Final Energy" in prop:
            return Units.convert(prop["DFTB Final Energy"], "au", unit)
        raise ResultsError("'DFTB Final Energy' not present in 'Properties' section of {}".format(self._kfpath()))

    def get_dipole_vector(self, unit="au"):
        """get_dipole_vector(unit='au')
        Return the dipole vector, expressed in *unit*.
        """
        prop = self.get_properties()
        if "Electric Dipole Moment" in prop:
            return Units.convert(prop["Electric Dipole Moment"], "au", unit)
        raise ResultsError("'Electric Dipole Moment' not present in 'Properties' section of {}".format(self._kfpath()))

    def get_gradients(self):
        """get_gradients()
        Return the list of atomic gradients, expressed in atomic units. Returned value is a list of vectors (lists of lenght 3).
        """
        prop = self.get_properties()
        if "Generic Gradient" in prop:
            grad = prop["Generic Gradient"]
            return [grad[i : i + 3] for i in range(0, len(grad), 3)]
        raise ResultsError("'Generic Gradient' not present in 'Properties' section of {}".format(self._kfpath()))

    def _int2inp(self):
        """_int2inp()
        In DFTB the internal order is always the same as the input order. Return an identity permutation of length equal to the number of atoms.
        """
        return list(range(1, 1 + self.readkf("Molecule", "nAtoms")))

    def _atomic_numbers_input_order(self):
        """_atomic_numbers_input_order()
        Return a list of atomic numbers, in the input order.
        """
        return self.readkf("Molecule", "AtomicNumbers")


class DFTBJob(SCMJob):
    _result_type = DFTBResults
    _command = "dftb"
    _top = ["units", "task"]
    _subblock_end = "end"

    def _serialize_mol(self):
        s = self.settings.input

        if len(self.molecule.lattice) in [1, 2] and self.molecule.align_lattice():
            log(
                "The lattice supplied for job {} did not follow the convention required by DFTB. I rotated the whole system for you. You're welcome".format(
                    self._full_name()
                ),
                3,
            )

        for i, atom in enumerate(self.molecule):
            s.system.atoms["_" + str(i + 1)] = atom.str(symbol=self._atom_symbol(atom), space=18, decimal=10)

        if self.molecule.lattice:
            for i, vec in enumerate(self.molecule.lattice):
                s.system.lattice["_" + str(i + 1)] = "{:16.10f} {:16.10f} {:16.10f}".format(*vec)

    def _remove_mol(self):
        s = self.settings.input
        if "system" in s:
            if "atoms" in s.system:
                del s.system.atoms
            if "lattice" in s.system:
                del s.system.lattice



from scm.plams import PT, Molecule
import pytest


PT.set_connectors("Mg", 4)


@pytest.fixture
def m1(xyz_folder):
    return Molecule(xyz_folder / "chlorophyl1.xyz")


@pytest.fixture
def m2(xyz_folder):
    return Molecule(xyz_folder / "chlorophyl2.xyz")


def testYES(m1, m2):
    for i in range(2):
        assert m1.label(i) == m2.label(i)


def testNO(m1, m2):
    for i in range(2, 5):
        assert m1.label(i) != m2.label(i)



#!/usr/bin/env python

from scm.plams.core.errors import PlamsError
import numpy
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.kftools import KFFile
from scm.plams.tools.periodic_table import PeriodicTable
from scm.plams.tools.units import Units
from scm.plams.trajectories.trajectoryfile import TrajectoryFile

__all__ = ["RKFTrajectoryFile", "write_general_section", "write_molecule_section"]

bohr_to_angstrom = Units.conversion_ratio("bohr", "angstrom")


class RKFTrajectoryFile(TrajectoryFile):
    """
    Class representing an RKF file containing a molecular trajectory

    An instance of this class has the following attributes:

    *   ``file_object`` -- A PLAMS |KFFile| object, referring to the actual RKF file
    *   ``position``    -- The frame to which the cursor is currently pointing in the RKF file
    *   ``mode``        -- Designates whether the file is in read or write mode ('rb' or 'wb')
    *   ``ntap``        -- The number of atoms in the molecular system (needs to be constant throughout)
    *   ``elements``    -- The elements of the atoms in the system (needs to be constant throughout)
    *   ``conect``      -- The connectivity information of the current frame
    *   ``mddata``      -- Read mode only: A dictionary containing data from the MDHistory section in the RKF file
    *   ``read_lattice``-- Read mode only: Wether the lattice vectors will be read from the file
    *   ``read_bonds``  -- Wether the connectivity information will be read from the file
    *   ``saving_freq`` -- How often the 'wb' file is written (default: only when :meth:`close` is called)

    An |RKFTrajectoryFile| object behaves very similar to a regular file object.
    It has read and write methods (:meth:`read_next` and :meth:`write_next`)
    that read and write from/to the position of the cursor in the ``file_object`` attribute.
    If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
    the cursor to any frame in the file and reads from there.
    The amount of information stored in memory is kept to a minimum, as only information from the latest frame
    is ever stored.

    Reading and writing to and from the files can be done as follows::

        >>> from scm.plams import RKFTrajectoryFile

        >>> rkf = RKFTrajectoryFile('ams.rkf')
        >>> mol = rkf.get_plamsmol()

        >>> rkfout = RKFTrajectoryFile('new.rkf',mode='wb')

        >>> for i in range(rkf.get_length()) :
        >>>     crd,cell = rkf.read_frame(i,molecule=mol)
        >>>     rkfout.write_next(molecule=mol)
        >>> rkfout.close()

    The above script reads information from the RKF file ``ams.rkf`` into the |Molecule| object ``mol``
    in a step-by-step manner.
    The |Molecule| object is then passed to the :meth:`write_next` method of the new |RKFTrajectoryFile|
    object corresponding to the new rkf file ``new.rkf``.

    The exact same result can also be achieved by iterating over the instance as a callable

        >>> rkf = RKFTrajectoryFile('ams.rkf')
        >>> mol = rkf.get_plamsmol()

        >>> rkfout = RKFTrajectoryFile('new.rkf',mode='wb')

        >>> for crd,cell in rkf(mol) :
        >>>     rkfout.write_next(molecule=mol)
        >>> rkfout.close()

    This procedure requires all coordinate information to be passed to and from the |Molecule| object
    for each frame, which can be time-consuming.
    Some time can be saved by bypassing the |Molecule| object::

        >>> rkf = RKFTrajectoryFile('ams.rkf')

        >>> rkfout = RKFTrajectoryFile('new.rkf',mode='wb')
        >>> rkfout.set_elements(rkf.get_elements())

        >>> for crd,cell in rkf :
        >>>     rkfout.write_next(coords=crd,cell=cell,conect=rkf.conect)
        >>> rkfout.close()

    The only mandatory argument to the :meth:`write_next` method is ``coords``.
    Further time can be saved by setting the ``read_lattice`` and ``read_bonds`` variables to False.

    By default the write mode will create a minimal version of the RKF file, containing only elements,
    coordinates, lattice, and connectivity information.
    This minimal file format can be read by AMSMovie.

    It is possible to store additional information, such as energies, velocities, and charges.
    To enable this, the method :meth:`store_mddata` needs to be called after creation,
    and a dictionary of mddata needs to be passed to the :meth:`write_next` method.
    When that is done, the AMS trajectory analysis tools can be used on the file.
    Restarting an MD run with such a file is however currently not possible::

        >>> rkf = RKFTrajectoryFile('ams.rkf')
        >>> rkf.store_mddata()
        >>> mol = rkf.get_plamsmol()

        >>> rkf_out = RKFTrajectoryFile('new.rkf',mode='wb')
        >>> rkf_out.store_mddata(rkf)

        >>> for i in range(len(rkf)) :
        >>>         crd,cell = rkf.read_frame(i,molecule=mol)
        >>>         rkf_out.write_next(molecule=mol,mddata=rkf.mddata)
        >>> rkf_out.close()
    """

    def __init__(self, filename, mode="rb", fileobject=None, ntap=None):
        """
        Initiates an RKFTrajectoryFile object

        * ``filename``   -- The path to the RKF file
        * ``mode``       -- The mode in which to open the RKF file ('rb' or 'wb')
        * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
        * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
        """
        # TODO: If the mddata option is set to True, then the file created here works with AMSMovie and the analysis tools.
        #      To also make is work for restarts, two things have to be added:
        #      1. The final velocities have to be converted from bohr/fs to bohr/au (1/41.341373336493)
        #         and stored in MDResuts%EndVelocities
        #      2. The final coordinates need to be copied to the Molecule section.

        self.position = 0
        if filename is not None:
            # fileobject = KFFile(filename,autosave=False,keep_file_open=True)
            fileobject = KFFile(filename, autosave=False)
            # fileobject = KFFile(filename,autosave=False,fastsave=True)
            # This fastsave option (no copying) was not worth it, so I removed it.
            if fileobject is None:
                raise PlamsError("KFFile %s not found." % (filename))
        self.file_object = fileobject
        self.mode = mode

        self.ntap = 0
        if ntap is not None:
            self.ntap = ntap
        self.firsttime = True
        self.coords = numpy.zeros((self.ntap, 3))  # Only for reading purposes,
        # to avoid creating the array each time
        # PLAMS molecule related settings
        self.elements = ["H"] * self.ntap
        self.current_molecule = None
        self.store_molecule = True  # Even if True, the molecule attribute is only stored during iteration

        # RKF specific attributes
        self.program = "trajectory"
        self.nvecs = 3
        self.latticevecs = numpy.zeros((3, 3))
        self.read_lattice = True  # Reading time can be saved by skipping the lattice info
        self.read_bonds = True
        self.cell = numpy.zeros((3, 3))
        self.conect = None
        self.timestep = None
        self.saving_freq = None  # By default the 'wb' file is only written upon closing
        # Saving more often is much slower.
        self.include_mddata = False
        self.mdhistory_name = "MDHistory"
        self.mddata = None
        self.mdunits = None
        self.mdblocksize = None
        self.mditems = None
        self.mdblockitems = None
        self._mdblock = {}
        self.include_historydata = False  # Any additional data along the history section will be stored
        self.historydata = None
        self.historyitems = None

        # Skip to the trajectory part of the file (only if in read mode, because coords are required in header)
        if self.mode == "rb":
            self._read_header()
        elif self.mode == "wb":
            new_sections = ["InputMolecule", "Molecule", "History", self.mdhistory_name, "MDResults"]
            new_sections += ["SystemVersionHistory"]
            sections = self.file_object.sections()
            if len(sections) > 0:
                for secname in sections:
                    if secname in new_sections or "ChemicalSystem" in secname:
                        self.file_object.delete_section(secname)
            #       raise PlamsError ('RKF file %s already exists'%(filename))
        elif self.mode == "ab":
            self._move_cursor_to_append_pos()
        else:
            raise PlamsError('Mode %s is invalid. Only "rb" and "wb" are allowed.' % (self.mode))

    def store_mddata(self, rkf=None):
        """
        Read/write an MDHistory section

        * ``rkf`` -- If in write mode an RKFTrajectoryFile object in read mode needs to be passed to extract unit info
        """
        self.include_mddata = True
        if "r" in self.mode or "a" in self.mode:
            self._set_mddata_units()
            self._set_mddata_items()
        elif "w" in self.mode:
            if rkf is not None:
                self.timestep = rkf.timestep
                self._set_mdunits(rkf.mdunits)

    def store_historydata(self):
        """
        Read/write non-standard entries in the History section
        """
        self.include_historydata = True

        if "r" in self.mode:
            # Set the History items
            sections = self.file_object.get_skeleton()
            section = "History"
            item_keys = [kn for kn in sections[section] if "ItemName" in kn]
            items = [self.file_object.read(section, kn) for kn in item_keys]
            standard_items = [
                "Coords",
                "nLatticeVectors",
                "LatticeVectors",
                "Bonds.Index",
                "Bonds.Atoms",
                "Bonds.Orders",
            ]
            self.historyitems = [item for item in items if not item in standard_items]

    def close(self, override_molecule_section_with_last_frame=True):
        """
        Execute all prior commands and cleanly close and garbage collect the RKF file
        """
        # Write the step info
        if self.timestep is not None and self.mode == "wb":
            self.file_object.write("MDResults", "StartStep", 0)
            self.file_object.write("MDResults", "StartTime[fs]", 0.0)
            nsteps = self.get_length()
            self.file_object.write("MDResults", "EndStep", nsteps - 1)
            self.file_object.write("MDResults", "EndTime[fs]", (nsteps - 1) * self.timestep)

        # Write to file
        if self.mode == "wb":
            if override_molecule_section_with_last_frame:
                # First write the last frame into the Molecule section
                self._rewrite_molecule()
            # Then write to file
            self.file_object.save()
        del self

    def _rewrite_molecule(self):
        """
        Overwrite the molecule section with the latest frame
        """
        molecule = self.get_plamsmol()
        crd, cell = self.read_last_frame(molecule=molecule)
        self._write_molecule_section(crd, cell, molecule=molecule)

    def _read_header(self, molecule_section="Molecule"):
        """
        Set up info required for reading frames
        """
        if not molecule_section in self.file_object:
            return
        self.elements = self.file_object.read(molecule_section, "AtomSymbols")
        # If read from memory and not from file (write mode), it is already a list
        if isinstance(self.elements, str):
            self.elements = self.elements.split()
        self.elements = [el.split(".")[0] for el in self.elements]
        if (self.mdhistory_name, "Time(1)") in self.file_object:
            times = self.file_object.read(self.mdhistory_name, "Time(1)")
            if isinstance(times, list):
                self.timestep = times[1]
        self.ntap = len(self.elements)
        self.coords = numpy.zeros((self.ntap, 3))

        # Set the lattice info
        if (molecule_section, "LatticeVectors") in self.file_object:
            self.latticevecs = numpy.array(self.file_object.read(molecule_section, "LatticeVectors"))
            self.nvecs = int(len(self.latticevecs) / 3)  # Why did I remove this line locally?!
            self.latticevecs = self.latticevecs.reshape((self.nvecs, 3))

    def _set_mddata_units(self):
        """
        Get the units for the mddata, if those are to be read
        """
        # Look for the items
        section = self.mdhistory_name
        sections = self.file_object.get_skeleton()
        if not self.mdhistory_name in sections:
            self.mdunits = {}
            return
        item_keys = [kn for kn in sections[section] if "ItemName" in kn]
        items = [self.file_object.read(section, kn) for kn in item_keys]

        # Get the data for each item
        unit_dic = {}
        for item in items:
            if "%s(units)" % (item) in self.file_object.get_skeleton()[section]:
                unit_dic[item] = self.file_object.read(section, "%s(units)" % (item))

        self.mdunits = unit_dic

    def _set_mddata_items(self):
        """
        Get all the items for the mddatam if those are to be read
        """
        sections = self.file_object.get_skeleton()
        section = self.mdhistory_name
        if not self.mdhistory_name in sections:
            self.mditems = []
            self.mdblockitems = []
            self.mdblocksize = 100
            return
        blocksize = self.file_object.read(section, "blockSize")
        item_keys = [kn for kn in sections[section] if "ItemName" in kn]
        items = [self.file_object.read(section, kn) for kn in item_keys]
        blockitems = []
        for item in items:
            dim = self.file_object.read(section, "%s(dim)" % (item))
            if dim == 1 and not self.file_object.read(section, "%s(perAtom)" % (item)):
                is_blockitem = True
                if (section, "%s(1)" % (item)) in self.file_object:
                    if isinstance(self.file_object.read(section, "%s(1)" % (item)), str):
                        is_blockitem = False
                if is_blockitem:
                    blockitems.append(item)
        items = [item for item in items if not item in blockitems]
        self.mdblocksize = blocksize
        self.mditems = items
        self.mdblockitems = blockitems

    def _move_cursor_to_append_pos(self):
        """
        Get the instance ready for appending
        """
        self._read_header()
        self.position = self.get_length()

    def _write_header(self, coords, cell, molecule=None):
        """
        Write Molecule info to file (elements, periodicity)
        """
        # First write the general section
        if "General" not in self.file_object:
            write_general_section(self.file_object, self.program)

        # Then write the input molecule
        self._update_celldata(cell)
        self._write_molecule_section(coords, cell, molecule=molecule)
        self._write_molecule_section(coords, cell, section="InputMolecule", molecule=molecule)
        if self.include_mddata:
            # Start setting up the MDHistory section as well
            self.mdblocksize = 100
            self.file_object.write(self.mdhistory_name, "blockSize", 100)

        # Now make sure that it is possible to read from the file as well
        self._read_header()

    def _update_celldata(self, cell):
        """
        Use the newly supplied cell to update the dimensionality of the system
        """
        shape = numpy.array(cell).shape
        if len(shape) == 2:
            self.nvecs = shape[0]
            self.latticevecs = numpy.zeros((self.nvecs, 3))  # Not really necessay

    def _write_molecule_section(self, coords, cell, section="Molecule", molecule=None):
        """
        Write the molecule section
        """
        write_molecule_section(self.file_object, coords, cell, self.elements, section, molecule)

    def _set_mdunits(self, mdunits):
        """
        Store the dictionary with MD Units
        """
        if self.include_mddata:
            self.mdunits = mdunits

    def get_plamsmol(self):
        """
        Extracts a PLAMS molecule object from the RKF file
        """
        section_dict = self.file_object.read_section("InputMolecule")
        if len(section_dict) == 0:
            section_dict = self.file_object.read_section("Molecule")
        plamsmol = Molecule._mol_from_rkf_section(section_dict)
        return plamsmol

    def read_frame(self, i, molecule=None):
        """
        Reads the relevant info from frame ``i`` and returns it, or stores it in ``molecule``

        * ``i``        -- The frame number to be read from the RKF file
        * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
        """
        # Read the cell data
        cell = None
        if self.read_lattice:
            try:
                cell = self._read_cell_data(i)
            except (KeyError, AttributeError):
                pass

        # Read the bond data
        conect = None
        if self.read_bonds:
            conect = self._read_bond_data(section="History", step=i)
        self.conect = conect

        # Read the coordinates, and possible pass them to molecule
        try:
            self._read_coordinates(i, molecule, cell)
            # This has changed self.coords behind the scenes
        except (KeyError, AttributeError):
            return None, None

        # Read and store any additional data in the history section
        if self.include_historydata:
            self._store_historydata_for_step(i)
        # Read and store all MDData for this frame
        try:
            if self.include_mddata:
                self._store_mddata_for_step(i)
        except AttributeError:  # this is triggered when self.file_object is None triggered via self.close()
            pass
        # Finalize
        if self.firsttime:
            self.firsttime = False

        self.position = i
        return self.coords, cell

    def _read_coordinates(self, i, molecule, cell):
        """
        Read the coordinates at step i, and possible pass them to molecule
        """
        if not self.coords.shape == (self.ntap, 3):
            raise PlamsError("coords attribute has been changed outside the class")
        coords = self.coords.reshape(self.ntap * 3)
        coords[:] = self.file_object.read("History", "Coords(%i)" % (i + 1))
        coords *= bohr_to_angstrom
        # This has changed self.coords behind the scenes

        # Create the molecule
        if isinstance(molecule, Molecule):
            cell_reduced = None
            if cell is not None:
                cell_reduced = cell[: self.nvecs]
            self._set_plamsmol(self.coords, cell_reduced, molecule)

    def _read_cell_data(self, i):
        """
        Read the cell data at step i
        """
        if not ("History", "LatticeVectors(%i)" % (i + 1)) in self.file_object:
            return None
        latticevecs = self.latticevecs.reshape(self.nvecs * 3)
        latticevecs[:] = self.file_object.read("History", "LatticeVectors(%i)" % (i + 1))  # * bohr_to_angstrom
        latticevecs *= bohr_to_angstrom
        # This changed self.latticevecs behind the scenes
        # self.cell[:self.nvecs] = latticevecs
        self.cell[: self.nvecs] = self.latticevecs
        cell = self.cell
        return cell

    def _read_bond_data(self, section, step=None):
        """
        Read the bond data from the rkf file
        """
        conect = None
        try:
            step_txt = ""
            if step is not None:
                step_txt = "(%i)" % (step + 1)
            if not ("History", "Bonds.Index%s" % (step_txt)) in self.file_object:
                return conect
            indices = self.file_object.read(section, "Bonds.Index%s" % (step_txt))
            connection_table = self.file_object.read(section, "Bonds.Atoms%s" % (step_txt))
            if isinstance(connection_table, int):
                connection_table = [connection_table]
            bond_orders = self.file_object.read(section, "Bonds.Orders%s" % (step_txt))
            if isinstance(bond_orders, float):
                bond_orders = [bond_orders]
            conect = {}
            for i, (start, end) in enumerate(zip(indices[:-1], indices[1:])):
                if end - start > 0:
                    # conect[i+1] = connection_table[start-1:end-1]
                    conect[i + 1] = [
                        (ia, o)
                        for ia, o in zip(connection_table[start - 1 : end - 1], bond_orders[start - 1 : end - 1])
                    ]
            # Now correct the connection table
            # conect = self.symmetrize_conect(conect)
        except (KeyError, AttributeError):
            pass
        return conect

    def get_regular_connection_table(self):
        """
        Get the connection table without the bond orders
        """
        conect = self.symmetrize_conect(self.conect)  # noqa, ToDo: review, return value or remove method?

    def _store_mddata_for_step(self, istep):
        """
        Store the data from the MDHistory section
        """
        if "w" in self.mode:
            return
        self.mddata = {}

        # Get the data for each item
        section = self.mdhistory_name
        for item in self.mditems:
            if (section, "%s(%i)" % (item, istep + 1)) in self.file_object:
                self.mddata[item] = self.file_object.read(section, "%s(%i)" % (item, istep + 1))
        for item in self.mdblockitems:
            block = int(istep / self.mdblocksize)
            pos = istep % self.mdblocksize
            # First check if this block was already stored in memory
            if item in self._mdblock:
                if block + 1 in self._mdblock[item]:
                    values = self._mdblock[item][block + 1]
                    self.mddata[item] = values[pos]
                    continue
            # If not, try to read the block
            if not (section, "%s(%i)" % (item, block + 1)) in self.file_object:
                continue
            values = self.file_object.read(section, "%s(%i)" % (item, block + 1))
            if isinstance(values, str):
                values = values.split()
            if not isinstance(values, list):
                values = [values]
            self._mdblock[item] = {block + 1: values}
            self.mddata[item] = values[pos]

    def _store_historydata_for_step(self, istep):
        """
        Store the extra data from the History section

        Note: Block format is not used in the History section
        """
        if "w" in self.mode:
            return
        if self.historydata is None:
            self.historydata = {}
        section = "History"
        for item in self.historyitems:
            if (section, "%s(%i)" % (item, istep + 1)) in self.file_object:
                self.historydata[item] = self.file_object.read(section, "%s(%i)" % (item, istep + 1))

    def _is_endoffile(self):
        """
        Reads and checks If the end of file is reached.
        """
        return ("History", "Coords(%i)" % (self.position + 1)) in self.file_object

    def read_next(self, molecule=None, read=True):
        """
        Reads coordinates and lattice vectors from the current position of the cursor and returns it

        * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
        * ``read``     -- If set to False the cursor will move to the next frame without reading
        """
        if not read and not self.firsttime:
            return self._move_cursor_without_reading()

        if self.firsttime:
            self.firsttime = False

        crd, vecs = self.read_frame(self.position, molecule)
        self.position += 1
        return crd, vecs

    def write_next(self, coords=None, molecule=None, cell=[0.0, 0.0, 0.0], conect=None, historydata=None, mddata=None):
        """
        Write frame to next position in trajectory file

        * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates in angstrom
        * ``molecule`` -- A molecule object to read the molecular data from
        * ``cell``     -- A set of lattice vectors (or cell diameters for an orthorhombic system) in angstrom
        * ``conect``   -- A dictionary containing the connectivity info (e.g. {1:[2],2:[1]})
        * ``historydata`` -- A dictionary containing additional variables to be written to the History section
        * ``mddata``   -- A dictionary containing the variables to be written to the MDHistory section

        The ``mddata`` dictionary can contain the following keys:
        ('TotalEnergy', 'PotentialEnergy', 'Step', 'Velocities', 'KineticEnergy',
        'Charges', 'ConservedEnergy', 'Time', 'Temperature')

        The ``historydata`` dictionary can contain for example:
        ('Energy','Gradients','StressTensor')
        All values must be in atomic units
        Numpy arrays or lists of lists will be flattened before they are written to the file

        .. note::

                Either ``coords`` or ``molecule`` are mandatory arguments
        """
        # Check for common error in the arguments
        if coords is not None:
            if isinstance(coords, Molecule):
                raise PlamsError("The PLAMS molecule needs to be passed as the second argument (molecule)")

        if isinstance(molecule, Molecule):
            coords, cell, elements, conect, _ = self._read_plamsmol(molecule, read_props=False)
            if self.position == 0:
                self.elements = elements
        # Make sure that the cell consists of vectors
        cell = self._convert_cell(cell)
        if conect is not None:
            if len(conect) == 0:
                conect = None
        self.conect = conect

        # Include a check on the size of coords?
        if len(coords) != len(self.elements):
            raise PlamsError("The coordinates do not match the rest of the trajectory")

        # If this is the first step, write the header
        if self.position == 0:
            self._write_header(coords, cell, molecule)
            self.firsttime = False

        # Define some local variables
        step = self.position
        if mddata is not None:
            if "Step" in mddata:
                step = mddata["Step"]
        # Energy should be read from mddata first, otherwise from historydata, otherwise set to zero
        energy = self._set_energy(mddata, historydata)
        if not self.include_historydata or historydata is None:
            historydata = {}
        historydata["Energy"] = energy

        # Write the history section
        counter = 1
        counter = self._write_history_entry(step, coords, cell, conect, historydata, counter)

        if self.include_mddata and mddata is not None:
            self._write_mdhistory_entry(mddata)

        self.position += 1

        if self.saving_freq is not None:
            if self.position % self.saving_freq == 0:
                self.file_object.save()

    def _set_energy(self, mddata, historydata):
        """
        Looks if an energy is passed as input, and it not, sets to zero
        """
        energy = None
        if mddata is not None:
            if "PotentialEnergy" in mddata:
                energy = mddata["PotentialEnergy"]
        if energy is None:
            if historydata is not None:
                if "Energy" in historydata:
                    energy = historydata["Energy"]
        if energy is None:
            energy = 0.0
        return energy

    def _write_history_entry(self, step, coords, cell, conect, historydata=None, counter=1):
        """
        Write the full entry into the History section
        """
        self.file_object.write("History", "nEntries", self.position + 1)
        self.file_object.write("History", "currentEntryOpen", False)
        self._write_keydata_in_history("Step", counter, False, 1, self.position + 1, step)
        counter += 1
        crd = [float(c) / bohr_to_angstrom for coord in coords for c in coord]
        self._write_keydata_in_history("Coords", counter, True, 3, self.position + 1, crd)
        counter += 1
        # self._write_keydata_in_history('Energy', counter, False, 1, self.position+1, energy)
        # counter += 1
        if cell is not None:
            self._write_keydata_in_history("nLatticeVectors", counter, False, 1, self.position + 1, self.nvecs)
            counter += 1
            vecs = [float(v) / bohr_to_angstrom for vec in cell for v in vec]
            # I should probably rethink the dimension of the lattice vectors (generalize it)
            self._write_keydata_in_history("LatticeVectors", counter, False, [3, 3], self.position + 1, vecs)
            counter += 1

        if historydata is not None:
            counter = self._write_dictionary_to_history(historydata, "History", counter)
        # if gradients is not None :
        #        grd = [float(g) for grad in gradients for g in grad]
        #        self._write_keydata_in_history('Gradients', counter, True, 3, self.position+1, grd)

        # if stresstensor is not None :
        #        stre = [float(s) for stress in stresstensor for s in stress]
        #        self._write_keydata_in_history('StressTensor', counter, False, [3,3], self.position+1, stre)

        # Write the bond info
        if conect is not None:
            counter = self._write_bonds_in_history(conect, counter, len(coords))

        return counter

    def _write_bonds_in_history(self, conect, counter, nats):
        """
        Write the bond data into the history section
        """
        # Get the bond orders out of the connection table
        connections = {}
        orders = {}
        for k in conect.keys():
            connections[k] = [t[0] if isinstance(t, tuple) else t for t in conect[k]]
            orders[k] = [t[1] if isinstance(t, tuple) else 1.0 for t in conect[k]]

        numbonds = 0
        indices = [1]
        connection_table = []
        bond_orders = []
        for iat in range(1, nats + 1):
            connections = []
            if iat in conect:
                connections = conect[iat]
            neighbors = []
            bos = []
            for t in connections:
                jat = t
                bo = 1.0
                if isinstance(t, tuple):
                    jat = t[0]
                    bo = t[1]
                # Correct for double counting
                if jat <= iat:
                    continue
                neighbors.append(jat)
                bos.append(bo)
            numbonds += len(bos)
            indices.append(numbonds + 1)
            connection_table += neighbors
            bond_orders += bos

        self.file_object.write("History", "Bonds.Index(%i)" % (self.position + 1), indices)
        self.file_object.write("History", "ItemName(%i)" % (counter), "%s" % ("Bonds.Index"))
        counter += 1

        self.file_object.write("History", "Bonds.Atoms(%i)" % (self.position + 1), connection_table)
        self.file_object.write("History", "ItemName(%i)" % (counter), "%s" % ("Bonds.Atoms"))
        counter += 1

        self.file_object.write("History", "Bonds.Orders(%i)" % (self.position + 1), bond_orders)
        self.file_object.write("History", "ItemName(%i)" % (counter), "%s" % ("Bonds.Orders"))
        counter += 1

        return counter

    def _write_mdhistory_entry(self, mddata):
        """
        Write the entry in the MDHistory section
        """
        counter = 1
        counter = self._write_dictionary_to_history(mddata, self.mdhistory_name, counter)

    def _write_dictionary_to_history(self, data, section, counter=1):
        """
        Add the entries of a dictionary to a History section
        """
        self.file_object.write(section, "nEntries", self.position + 1)
        self.file_object.write(section, "currentEntryOpen", False)
        for key, var in data.items():
            # Make sure that the entry is either a scalar or a 1D list
            var = self._flatten_variable(var)
            peratom = False
            dim = 1
            if isinstance(var, list):
                if len(var) % len(self.elements) == 0:
                    dim = int(len(var) / len(self.elements))
                    peratom = True
                else:
                    dim = len(var)
            self._write_keydata_in_history(key, counter, peratom, dim, self.position + 1, var, section)
            counter += 1
        return counter

    def _flatten_variable(self, var):
        """
        Make sure that the variable is a Python 1D list (not numpy)
        """
        while True:
            if isinstance(var, list) or isinstance(var, numpy.ndarray):
                if len(var) == 0:
                    break
                if isinstance(var[0], list) or isinstance(var[0], numpy.ndarray):
                    var = [v for varitem in var for v in varitem]
                else:
                    if isinstance(var[0], numpy.int64):
                        var = [int(v) for v in var]
                    elif isinstance(var[0], numpy.float64):
                        var = [float(v) for v in var]
                    break
            else:
                if isinstance(var, numpy.int64):
                    var = int(var)
                elif isinstance(var, numpy.float64):
                    var = float(var)
                break
        return var

    def _write_keydata_in_history(self, key, i, perAtom, dim, step, values, section="History"):
        """
        Write all data about a key value in KFFile
        """
        # Some data only needs to be printed once
        printstartdata = False
        if step == 1:
            printstartdata = True

        # Block code: if the data is to be written as blocks, then step and values need to be replaced.
        if section == self.mdhistory_name:
            step, values = self._get_block_info(key, perAtom, dim, step, values, section)

        # The rest should be independent on format (block or individual)
        self.file_object.write(section, "%s(%i)" % (key, step), values)
        if printstartdata:
            self.file_object.write(section, "ItemName(%i)" % (i), "%s" % (key))
            self.file_object.write(section, "%s(perAtom)" % (key), perAtom)
            self.file_object.write(section, "%s(dim)" % (key), dim)
            if section == self.mdhistory_name and self.mdunits is not None:
                if key in self.mdunits:
                    self.file_object.write(section, "%s(units)" % (key), self.mdunits[key])

    def _get_block_info(self, key, perAtom, dim, step, values, section):
        """
        If the data is to be written as blocks, then step and values need to be replaced.
        """
        if dim == 1 and not perAtom and not (isinstance(values, list) or isinstance(values, str)):
            if not self.include_mddata:
                raise Exception("Set include_mddata to write the MD section.")
            iblock = int((step - 1) / self.mdblocksize) + 1
            if step % self.mdblocksize != 1:
                old_values = []
                if key in self._mdblock:
                    if iblock in self._mdblock[key]:
                        if len(self._mdblock[key][iblock]) == step % self.mdblocksize - 1:
                            old_values = self._mdblock[key][iblock]
                if len(old_values) == 0:
                    if (section, "%s(%i)" % (key, iblock)) in self.file_object:
                        old_values = self.file_object.read(section, "%s(%i)" % (key, iblock))
                        if not isinstance(old_values, list):
                            old_values = [old_values]
                values = old_values + [values]  # Values is a scalar
            else:
                self.file_object.write(section, "nBlocks", iblock)
            step = iblock
            self._mdblock[key] = {iblock: values}
            if not isinstance(values, list):
                self._mdblock[key] = {iblock: [values]}
        return step, values

    def rewind(self, nframes=None):
        """
        Rewind the file either by ``nframes`` or to the first frame

        *   ``nframes`` -- The number of frames to rewind
        """
        self.firsttime = True
        self.position = 0

    def get_length(self):
        """
        Get the number of frames in the file
        """
        nsteps = 0
        if "History" in self.file_object:
            nsteps = self.file_object.read("History", "nEntries")
        return nsteps

    def read_last_frame(self, molecule=None):
        """
        Reads the last frame from the file
        """
        nsteps = self.get_length()
        crd, cell = self.read_frame(nsteps - 1, molecule)
        return crd, cell


def write_general_section(rkf, program="plams"):
    """
    Write the General section of the RKF file
    """
    rkf.write("General", "file-ident", "RKF")
    rkf.write("General", "termination status", "NORMAL TERMINATION")
    rkf.write("General", "program", "%s" % (program))
    rkf.write("General", "user input", " ")


def write_molecule_section(rkf, coords=None, cell=None, elements=None, section="Molecule", molecule=None):
    """
    Write the molecule section

    Note: Currently does not write bonds
    """
    if molecule is not None:
        if coords is None:
            coords = molecule.as_array()
        if cell is None and len(molecule.lattice) > 0:
            cell = molecule.lattice
        if elements is None:
            elements = [at.symbol for at in molecule.atoms]

    # Then write the input molecule
    charge = 0.0
    if molecule is not None:
        if "charge" in molecule.properties:
            charge = float(molecule.properties.charge)
    element_numbers = [PeriodicTable.get_atomic_number(el) for el in elements]

    rkf.write(section, "nAtoms", len(elements))
    rkf.write(section, "AtomicNumbers", element_numbers)
    rkf.write(section, "AtomSymbols", elements)
    crd = [Units.convert(float(c), "angstrom", "bohr") for coord in coords for c in coord]
    rkf.write(section, "Coords", crd)
    rkf.write(section, "Charge", charge)
    if cell is not None:
        rkf.write(section, "nLatticeVectors", len(cell))
        vecs = [Units.convert(float(v), "angstrom", "bohr") for vec in cell for v in vec]
        rkf.write(section, "LatticeVectors", vecs)
    # Should it write bonds?
    # Write atom properties
    if molecule is not None:
        from scm.plams.interfaces.adfsuite.ams import AMSJob

        suffixes = [AMSJob._atom_suffix(at) for at in molecule]
        if any(s != "" for s in suffixes):
            rkf.write(section, "EngineAtomicInfo", "\x00".join(suffixes))
        # Add atomic charges
        charges = [at.properties.forcefield for at in molecule.atoms if "forcefield" in at.properties.keys()]
        charges = [float(s.charge) for s in charges if "charge" in s.keys()]
        if len(charges) == len(molecule):
            rkf.write(section, "Charges", charges)
        # Add region sections
        if "regions" in molecule.properties.keys():
            region_names = molecule.properties.regions.keys()
            rkf.write(section, "RegionNames", "\x00".join(region_names))
            rkf.write(
                section,
                "RegionProperties",
                "\x00".join(["\n".join(molecule.properties.regions[k]) for k in region_names]),
            )
        # Also add a bond section
        if len(molecule.bonds) > 0:
            bond_indices = [sorted([iat for iat in molecule.index(bond)]) for bond in molecule.bonds]
            atoms_from = [bond[0] for bond in bond_indices]
            atoms_to = [bond[1] for bond in bond_indices]
            orders = [float(bond.order) for bond in molecule.bonds]
            rkf.write(section, "fromAtoms", atoms_from)
            rkf.write(section, "toAtoms", atoms_to)
            rkf.write(section, "bondOrders", orders)

            # I also need to write the lattice displacements of the bonds
            # To do that, I need to compute them.
            # I could make a start with writing in zeros
            if cell is not None:
                lattice_displacements = compute_lattice_displacements(molecule)
                # lattice_displacements = [0 for i in range(len(cell)*len(molecule.bonds))]
                rkf.write(section, "latticeDisplacements", lattice_displacements)


def compute_lattice_displacements(molecule):
    """
    Determine which bonds are displaced along the periodic lattice, so that they are not at their closest distance
    """
    cell = numpy.array(molecule.lattice)
    nvecs = len(cell)

    # Get the difference vectors for the bonds
    nbonds = len(molecule.bonds)
    bond_indices = numpy.array([sorted([iat - 1 for iat in molecule.index(bond)]) for bond in molecule.bonds])
    coords = molecule.as_array()
    vectors = coords[bond_indices[:, 0]] - coords[bond_indices[:, 1]]

    # Project the vectors onto the lattice vectors
    celldiameters_sqrd = (cell**2).sum(axis=1)
    proj = (vectors.reshape((nbonds, 1, 3)) * cell.reshape(1, nvecs, 3)).sum(axis=2)
    proj = proj / celldiameters_sqrd.reshape((1, nvecs))

    # Now see what multiple they are of 0.5
    lattice_displacements = numpy.round(proj).astype(int)
    lattice_displacements = lattice_displacements.reshape((nvecs * nbonds)).tolist()
    return lattice_displacements



# from .scmjob import SCMJob, SCMResults
import math

import numpy
from scm.plams.tools.kftools import KFFile

__all__ = ["FCFDOS"]


class FCFDOS:
    """
    A class for calculating the convolution of two FCF spectra
    """

    # Conversion factor to cm-1 == ( me^2 * c * alpha^2 ) / ( 100 * 2pi * amu * hbar )
    G2F = 120.399494933
    # J2Ha = 1.0 / (scipy.constants.m_e * scipy.constants.speed_of_light**2 * 0.0072973525693**2)
    J2Ha = 2.293712278400752e17

    def __init__(self, absrkf, emirkf, absele=0.0, emiele=0.0):
        """
        absrkf : KFFile from an FCF absorption calculation for the acceptor, as name, path, or KFFile object
        emirkf : KFFile from an FCF emission calculation for the donor, as name, path, or KFFile object
        absele : Acceptor absorption electronic energy cm-1
        emiele : Donor emission electronic energy in cm-1
        """
        if isinstance(absrkf, str):
            self.absrkf = KFFile(absrkf)
        else:
            self.absrkf = absrkf
        if isinstance(emirkf, str):
            self.emirkf = KFFile(emirkf)
        else:
            self.emirkf = emirkf
        self.absele = absele
        self.emiele = emiele
        self.spc = None

    def _getstick(self, source):
        spclen = source.read("Fcf", "nspectrum")
        rawspc = source.read("Fcf", "spectrum")
        stick = numpy.reshape(numpy.array(rawspc), (2, spclen)).transpose()
        # Reorder spectrum if decreasing
        if stick[0, 0] > stick[-1, 0]:
            for ix in range(numpy.size(stick, 0) // 2):
                XX = numpy.copy(stick[ix, :])
                stick[ix, :] = stick[-ix - 1, :]
                stick[-ix - 1, :] = XX
        # Get frequencies in cm-1
        frq1 = numpy.array(source.read("Fcf", "gamma1")) * self.G2F
        frq2 = numpy.array(source.read("Fcf", "gamma2")) * self.G2F
        # Calculate energy in Joules
        # factor = scipy.constants.pi * scipy.constants.hbar * scipy.constants.speed_of_light * 100.0
        factor = 9.932229285744643e-24
        viben1 = sum(frq1) * factor
        viben2 = sum(frq2) * factor
        # Convert to Hartree
        viben1 = viben1 * self.J2Ha
        viben2 = viben2 * self.J2Ha
        # Add ZPE and electronic energy
        stick[:, 0] = stick[:, 0] + viben2 - viben1 + self.absele + self.emiele
        return stick

    def _spcinit(self):
        # Get absorption and emission spectra
        absspc = self._getstick(self.absrkf)
        emispc = self._getstick(self.emirkf)
        # Find spectrum bounds
        absmin = numpy.amin(absspc[:, 0])
        absmax = numpy.amax(absspc[:, 0])
        emimin = numpy.amin(emispc[:, 0])
        emimax = numpy.amax(emispc[:, 0])
        spcmin = min(absmin, emimin)
        spcmax = max(absmax, emimax)
        # Find spectrum length and grain
        absdlt = abs(absspc[0, 0] - absspc[1, 0])
        emidlt = abs(absspc[0, 0] - absspc[1, 0])
        spcgrn = min(absdlt, emidlt)
        spclen = math.floor((spcmax - spcmin) / spcgrn) + 1
        # Initialize spectrum
        self.spc = self.newspc(spcmin, spcmax, spclen)
        return None

    def dos(self, lineshape="GAU", HWHM=100.0):
        """
        Calculate density of states by computing the overlap of the two FCF spectra
        The two spectra are broadened using the chosen lineshape and Half-Width at Half-Maximum in cm-1
        """
        # Initialize spectrum
        self._spcinit()
        # Get stick spectra
        absstick = self._getstick(self.absrkf)
        emistick = self._getstick(self.emirkf)
        # Convolute with gaussians
        absspc = numpy.copy(self.spc)
        emispc = numpy.copy(self.spc)
        absspc = self.convolute(absspc, absstick, lineshape, HWHM)
        emispc = self.convolute(emispc, emistick, lineshape, HWHM)
        # Integrate
        # Calculate DOS
        self.spc[:, 1] = absspc[:, 1] * emispc[:, 1]
        dos = self.trapezoid(self.spc)
        return dos

    def newspc(self, spcmin, spcmax, spclen=1000):
        # Dimensions of the spectrum
        delta = (spcmax - spcmin) / (spclen - 1)
        spc = numpy.zeros((spclen, 2), dtype=float)
        for ix in range(spclen):
            spc[ix, 0] = spcmin + delta * ix
        return spc

    def convolute(self, spc, stick, lineshape=None, HWHM=None):
        """
        Convolute stick spectrum with the chosen width and lineshape
        lineshape : Can be Gaussian or Lorentzian
        HWHM      : expressed in cm-1
        """
        if HWHM is None:
            raise ValueError("HWHM not defined")
        if lineshape is None:
            raise ValueError("Lineshape not defined")
        # Data for the convolution
        delta = spc[1, 0] - spc[0, 0]
        if lineshape[0:3].upper() == "GAU":
            # Gaussian lineshape
            idline = 1
            # This includes the Gaussian prefactor and the factor to account for the reduced lineshape width
            # factor = 1. / scipy.special.erf(2*math.sqrt(math.log(2)))
            factor = 1.0188815852036244
            factA = math.sqrt(numpy.log(2.0) / math.pi) * factor / HWHM
            factB = -numpy.log(2.0) / HWHM**2
            # We only convolute between -2HWHM and +2HWHM which accounts for 98.1% of the area
            ishft = math.floor(2 * HWHM / delta)
        elif lineshape[0:3].upper() == "LOR":
            # Lorentzian lineshape
            idline = 2
            # This includes the Lorentzian prefactor and the factor to account for the reduced lineshape width
            factA = (math.pi / math.atan(12) / 2) * HWHM / math.pi
            factB = 0.0
            # We only convolute between -12HWHM and +12HWHM which accounts for 94.7% of the area
            ishft = math.floor(12 * HWHM / delta)
        else:
            raise ValueError("invalid lineshape")
        # Loop over peaks in the stick spectrum
        spclen = numpy.size(spc, 0)
        for ix in range(numpy.size(stick[:, 0])):
            # Find peak position in the convoluted spectrum
            peakpos = 1 + math.floor((stick[ix, 0] - spc[0, 0]) / delta)
            # Convolution interval, limited to save computational time
            i1 = max(peakpos - ishft, 1)
            i2 = min(peakpos + ishft, spclen)
            factor = factA * stick[ix, 1]
            if idline == 1:  # Gaussian
                for i in range(i1, i2):
                    spc[i, 1] = spc[i, 1] + factor * math.exp(factB * (spc[i, 0] - stick[ix, 0]) ** 2)
            elif idline == 2:  # Lorentzian
                for i in range(i1, i2):
                    spc[i, 1] = spc[i, 1] + factor / (HWHM + (spc[i, 0] - stick[ix, 0]) ** 2)
        return spc

    def trapezoid(self, spc):
        """
        Integrate spectrum using the trapezoid rule
        """
        value = (spc[0, 1] + spc[-1, 1]) / 2
        value = value + sum(spc[1:-1, 1])
        # The abscissas must be equally spaced for this to work
        value = value * (spc[1, 0] - spc[0, 0])
        return value

    def __str__(self):
        string = f"Absorption from {self.absrkf.path}\nEmission from {self.emirkf.path}\nAcceptor absorption electronic energy = {self.absele} cm-1\nDonor emission electronic energy = {self.emiele} cm-1"
        return string



#!/usr/bin/env amspython

import sys

from scm.plams import AMSJob, Settings

# Read from a previous result file
path = sys.argv[1]
job = AMSJob.load_external(path)
charges, types, patch = job.results.get_forcefield_params()
print(charges)
print(types)
print(patch)

# Add the atom types and charges to the molecule
assert job.molecule is not None
mol = job.molecule.copy()
for i, at in enumerate(mol.atoms):
    at.properties.ForceField.Charge = charges[i]
    at.properties.ForceField.Type = types[i]

# Write a patch file
with open("patch.dat", "w") as outfile:
    outfile.write(str(patch))
    outfile.close()

# Set up new job settings
settings = Settings()
settings.input.ForceField.Type = "GAFF"
settings.input.ForceField.ForceFieldPatchFile = "patch.dat"
settings.input.ams.Task = "SinglePoint"

# Create the new job, and write the input file
job = AMSJob(molecule=mol, settings=settings)
text = job.get_input()
with open("ams.in", "w") as outfile:
    outfile.write(text)
    outfile.close()



#!/usr/bin/env amspython
from scm.plams import *
from scm.plams.recipes.redox import AMSRedoxScreeningJob, AMSRedoxDirectJob, AMSRedoxThermodynamicCycleJob


def main():
    mol = from_smiles("C1=CC(=O)C=CC1=O", forcefield="uff")  # benzoquinone
    solvent_name = "Water"  # Solvent for AMSRedoxDirectJob and AMSRedoxThermodynamicCycleJob. See the ADF Solvation documentation for which solvents are available.
    solvent_coskf = CRSJob.database() + "/Water.coskf"  # .coskf file or AMSRedoxScreeningJob

    vibrations = True  # set to False to turn off vibrations for AMSRedoxDirectJob and AMSRedoxThermodynamicCycleJob
    reduction = True
    oxidation = True  # not really relevant for benzoquinone, set to False to not calculate

    # DFT settings for AMSRedoxDirectJob and AMSRedoxThermodynamicCycleJob
    # Note: These are for demonstration purposes only. Use a better functional/settings for production purposes.
    s = Settings()
    s.input.adf.basis.type = "DZP"
    s.input.adf.xc.gga = "PBE"
    s.input.ams.GeometryOptimization.Convergence.Quality = "Basic"

    jobs = [
        AMSRedoxScreeningJob(
            name="quick_screening", molecule=mol, reduction=reduction, oxidation=oxidation, solvent_coskf=solvent_coskf
        ),
        AMSRedoxDirectJob(
            name="direct_best_method",
            molecule=mol,
            settings=s,
            reduction=reduction,
            oxidation=oxidation,
            vibrations=vibrations,
            solvent=solvent_name,
        ),
        AMSRedoxThermodynamicCycleJob(
            name="thermodynamic_cycle",
            molecule=mol,
            settings=s,
            reduction=reduction,
            oxidation=oxidation,
            vibrations=vibrations,
            solvent=solvent_name,
        ),
    ]

    for job in jobs:
        job.run()
        print_results([job])

    print("Final summary:")
    print_results(jobs)


def print_results(jobs):
    SHE = 4.42  # standard hydrogen electrode in eV on absolute scale

    print("The experimental reduction potential of benzoquinone is +0.10 V vs. SHE")
    print(
        "{:24s} {:24s} {:24s} {:24s} {:24s}".format(
            "Jobname", "Eox(vib,rel-to-SHE)[V]", "Ered(vib,rel-to-SHE)[V]", "Eox(rel-to-SHE)[V]", "Ered(rel-to-SHE)[V]"
        )
    )

    for job in jobs:
        s = f"{job.name:24s}"
        for vibrations in [True, False]:
            try:
                Eox = job.results.get_oxidation_potential(vibrations=vibrations) - SHE
                Eox = f"{Eox:.2f}"
            except:
                Eox = "N/A"
            s += f" {Eox:24s}"

            try:
                Ered = job.results.get_reduction_potential(vibrations=vibrations) - SHE
                Ered = f"{Ered:.2f}"
            except:
                Ered = "N/A"
            s += f" {Ered:24s}"

        print(s)


if __name__ == "__main__":
    init()
    main()
    finish()



#!/usr/bin/env amspython
import matplotlib.pyplot as plt
import pyCRS

"""

This example uses the Property Prediction tool to estimate some properties for
a molecule represented by a SMILES string.

The vapor pressures are handled separately from other properties because vapor pressures are temperature-dependent.

Run this script using
$AMSBIN/amspython PropertyPrediction.py

"""


def single_compound_example():
    print("Single compound example for SMILES CCO")
    mol = pyCRS.Input.read_smiles("CCO")
    temperatures = [298.15, 308.15, 318.15]
    pyCRS.PropPred.estimate(mol, temperatures=temperatures)
    print("Boiling point: {:.2f} {}".format(mol.properties["boilingpoint"], pyCRS.PropPred.units["boilingpoint"]))
    print("Available properties: {}".format(pyCRS.PropPred.available_properties))
    # vapor pressures are handled separately -- as part of temperature-dependent properties
    print(f'{"Temperature":>15} {"Vapor Pressure":>15}')
    for i in range(len(temperatures)):
        print(
            f'{mol.properties_tdep["vaporpressure"][i][0]:15.2f} {mol.properties_tdep["vaporpressure"][i][1]:15.6e} {pyCRS.PropPred.units["vaporpressure"]}'
        )


def multi_compound_example():
    smiles_list = ["CCO", "CCOC", "OCCCN", "C", "C1=CC=C(C=C1)COCC2=CC=CC=C2"]
    mols = [pyCRS.Input.read_smiles(s) for s in smiles_list]
    for mol in mols:
        pyCRS.PropPred.estimate(mol, temperatures=list(range(280, 340, 10)))

    plot_vapor_pressures(mols, "vaporpressures.png")


def property_list_example():
    print("Using a specified list of properties only")
    mol = pyCRS.Input.read_smiles("CCCCO")
    pyCRS.PropPred.estimate(
        mol, ["boilingpoint", "criticaltemp", "density", "vaporpressure"], temperatures=[290, 300, 310, 320, 330]
    )

    print("Temperature-independent properties:")
    for k, v in mol.properties.items():
        print(f"{k:<15} {v:15.6e} {pyCRS.PropPred.units[k]}")
    print("\nTemperature-dependent properties")
    for k, vals in mol.properties_tdep.items():
        print(k)
        print(f'{"Temperature":>15} {"Vapor Pressure":>15}')
        for temp, val in vals:
            print(f"{temp:15.2f} {val:15.6e} {pyCRS.PropPred.units[k]}")


def plot_vapor_pressures(mols, filename: str):
    n = 2  # only show the first two

    plt.clf()
    for mol in mols[:n]:
        temperatures = [x[0] for x in mol.properties_tdep["vaporpressure"]]
        vaporpressures = [x[1] for x in mol.properties_tdep["vaporpressure"]]
        plt.plot(temperatures, vaporpressures, label=mol.smiles)
    plt.legend()
    plt.xlabel("Temperature (K)")
    plt.ylabel("Vapor pressure (bar)")
    plt.title("Estimated vapor pressures")
    plt.plot()
    plt.savefig(filename)


if __name__ == "__main__":
    single_compound_example()
    multi_compound_example()
    property_list_example()



#!/usr/bin/env python

from scm.plams.core.errors import TrajectoryError
import numpy
from scm.plams.mol.molecule import Molecule
from scm.plams.tools.geometry import cell_shape, cellvectors_from_shape
from scm.plams.trajectories.trajectoryfile import TrajectoryFile

__all__ = ["XYZTrajectoryFile", "create_xyz_string"]


class XYZTrajectoryFile(TrajectoryFile):
    """
    Class representing an XYZ file containing a molecular trajectory

    An instance of this class has the following attributes:

    *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual XYZ file
    *   ``position``    -- The frame to which the cursor is currently pointing in the XYZ file
    *   ``mode``        -- Designates whether the file is in read or write mode ('r' or 'w')
    *   ``ntap``        -- The number of atoms in the molecular system (needs to be constant throughout)
    *   ``elements``    -- The elements of the atoms in the system (needs to be constant throughout)

    An |XYZTrajectoryFile| object behaves very similar to a regular file object.
    It has read and write methods (:meth:`read_next` and :meth:`write_next`)
    that read and write from/to the position of the cursor in the ``file_object`` attribute.
    If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
    the cursor to any frame in the file and reads from there.
    The amount of information stored in memory is kept to a minimum, as only information from the current frame
    is ever stored.

    Reading and writing to and from the files can be done as follows::

        >>> from scm.plams import XYZTrajectoryFile

        >>> xyz = XYZTrajectoryFile('old.xyz')
        >>> mol = xyz.get_plamsmol()

        >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')

        >>> for i in range(xyz.get_length()) :
        >>>     crd,cell = xyz.read_frame(i,molecule=mol)
        >>>     xyzout.write_next(molecule=mol)

    The above script reads information from the XYZ file ``old.xyz`` into the |Molecule| object ``mol``
    in a step-by-step manner.
    The |Molecule| object is then passed to the :meth:`write_next` method of the new |XYZTrajectoryFile|
    object corresponding to the new xyz file ``new.xyz``.

    The exact same result can also be achieved by iterating over the instance as a callable

        >>> xyz = XYZTrajectoryFile('old.xyz')
        >>> mol = xyz.get_plamsmol()

        >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')

        >>> for crd,cell in xyz(mol) :
        >>>     xyzout.write_next(molecule=mol)

    This procedure requires all coordinate information to be passed to and from the |Molecule| object
    for each frame, which can be time-consuming.
    It is therefore also possible to bypass the |Molecule| object when reading through the frames::

        >>> xyz = XYZTrajectoryFile('old.xyz')

        >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')
        >>> xyzout.set_elements(xyz.get_elements())

        >>> for crd,cell in xyz :
        >>>     xyzout.write_next(coords=crd)
        >>> xyzout.close()

    By default the write mode will create a minimal version of the XYZ file, containing only elements
    and coordinates.
    Additional information can be written to the file by supplying additional arguments
    to the :meth:`write_next` method.
    The additional keywords `step` and `energy` trigger the writing of a remark containing
    the molecule name, the step number, the energy, and the lattice vectors.

        >>> mol = Molecule('singleframe.xyz')

        >>> xyzout = XYZTrajectoryFile('new.xyz',mode='w')
        >>> xyzout.set_name('MyMol')

        >>> xyzout.write_next(molecule=mol, step=0, energy=5.)
    """

    def __init__(self, filename, mode="r", fileobject=None, ntap=None):
        """
        Initiates an XYZTrajectoryFile object

        * ``filename``   -- The path to the XYZ file
        * ``mode``       -- The mode in which to open the XYZ file ('r' or 'w')
        * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
        * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
        """
        TrajectoryFile.__init__(self, filename, mode, fileobject, ntap)

        # XYZ specific attributes
        self.name = "PlamsMol"

        # Specific XYZ stuff
        self.include_historydata = False
        self.historydata = None
        self.nveclines = 0

        # Required setup before frames can be read/written
        if self.mode == "r":
            self._read_header()
        elif self.mode == "a":
            self._move_cursor_to_append_pos()

    def store_historydata(self):
        """
        Additional data should be read from/written to file
        """
        self.include_historydata = True

    def set_name(self, name):
        """
        Sets the name of the system, in case an extensive write is requested

        *   ``name`` -- A string containing the name of the molecule
        """
        self.name = name

    def _read_header(self):
        """
        Set up info required for reading frames
        """
        line = self.file_object.readline()
        self.ntap = int(line.split()[0])
        if self.coords.shape == (0, 3):
            self.coords = numpy.zeros((self.ntap, 3))

        self.file_object.readline()

        elements = []
        for i in range(self.ntap):
            line = self.file_object.readline()
            elements.append(line.split()[0])
        self.elements = elements

        # See if there are any vector lines
        while True:
            line = self.file_object.readline()
            if len(line) > 0 and line.split()[0][:3] == "VEC":
                self.nveclines += 1
            else:
                break

        self.file_object.seek(0)

    def read_next(self, molecule=None, read=True):
        """
        Reads coordinates from the current position of the cursor and returns it

        * ``molecule`` -- |Molecule| object in which the new coordinates need to be stored
        * ``read``     -- If set to False the cursor will move to the next frame without reading
        """
        if not read and not self.firsttime:
            return self._move_cursor_without_reading()

        cell = None
        # Read the coordinates
        crd, cell = self._read_coordinates(molecule)
        if crd is None:
            return None, None  # End of file is reached

        if self.firsttime:
            self.firsttime = False

        self.position += 1

        return self.coords, cell

    def _read_coordinates(self, molecule):
        """
        Read the coordinates from file, and place them in the molecule
        """
        cell = None
        for i in range(2):
            line = self.file_object.readline()
            if i == 0 and len(line.split()) > 1:
                raise TrajectoryError("Number of atoms changes. Try XYZHistoryFile")
            elif i == 0 and int(line.split()[0]) != self.ntap:
                raise TrajectoryError("Number of atoms changes. Try XYZHistoryFile")
            if len(line) == 0:
                return None, None  # End of file is reached
        # Handle the comment line
        historydata = data_from_xyzcomment(line)
        if "Lattice" in historydata:
            cell = historydata["Lattice"]
            del historydata["Lattice"]
        if self.include_historydata:
            self.historydata = historydata

        # Reade coordinates
        for i in range(self.ntap):
            line = self.file_object.readline()
            self.coords[i, :] = [float(w) for w in line.split()[1:4]]

        # Possibly read lattice
        lattice = []
        for i in range(self.nveclines):
            line = self.file_object.readline()
            words = line.split()
            lattice.append([float(w) for w in words[1:]])
        if cell is None and len(lattice) > 0:
            cell = lattice

        if isinstance(molecule, Molecule):
            self._set_plamsmol(self.coords, cell, molecule)

        return self.coords, cell

    def _is_endoffile(self):
        """
        If the end of file is reached, return coords and cell as None
        """
        end = False
        for i in range(self.ntap + 2 + self.nveclines):
            line = self.file_object.readline()
            if len(line) == 0:
                end = True
                break
        return end

    def write_next(self, coords=None, molecule=None, cell=[0.0, 0.0, 0.0], conect=None, historydata=None):
        """
        Write frame to next position in trajectory file

        * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
        * ``molecule`` -- A molecule object to read the molecular data from
        * ``cell``     -- A set of lattice vectors or cell diameters
        * ``conect``   -- A dictionary containing connectivity info (not used)
        * ``historydata`` -- A dictionary containing additional variables to be written to the comment line

        The ``historydata`` dictionary can contain for example:
        ('Step','Energy'), the frame number and the energy respectively

        .. note::

                Either ``coords`` or ``molecule`` are mandatory arguments
        """
        if isinstance(molecule, Molecule):
            coords, cell, elements = self._read_plamsmol(molecule)[:3]
            if self.position == 0:
                self.elements = elements
        cell = self._convert_cell(cell)

        self._write_moldata(coords, cell, historydata)

        self.position += 1

    def _write_moldata(self, coords, cell, historydata):
        """
        Write all molecular info to file
        """
        if historydata is None:
            historydata = {}
        if self.include_historydata and len(historydata) > 0:
            step = self.position
            if "Step" in historydata:
                step = historydata["Step"]
            energy = 0.0
            if "Energy" in historydata:
                energy = historydata["Energy"]
            box = None
            if cell is not None:
                # box = PDBMolecule().box_from_vectors(cell)
                box = cell_shape(cell)
            name = self.name
            if "Name" in historydata:
                name = historydata["Name"]
            line = None
            if "Line" in historydata:
                line = historydata["Line"]
            block = create_xyz_string(self.elements, coords, energy, box, step, name, line)
        else:
            block = create_xyz_string(self.elements, coords)
        self.file_object.write(block)

    def _rewind_to_first_frame(self):
        """
        Rewind the file to the first frame
        """
        self.file_object.seek(0)
        self.firsttime = True
        self.position = 0

    def _rewind_n_frames(self, nframes):
        """
        Rewind the file by nframes frames
        """
        new_frame = self.position - nframes
        self._rewind_to_first_frame()
        for i in range(new_frame):
            self.read_next(read=False)


def create_xyz_string(elements, coords, energy=None, box=None, step=None, name="PlamsMol", line=None):
    """
    Write an XYZ file based on the elements and the coordinates of the atoms
    """
    block = "%i\n" % (len(elements))
    if line is not None:
        block += "%s" % (line)
    elif step is not None:
        if energy is None:
            energy = 0.0
        comment = "%-40s%6i %16.6f" % (name, step, energy)
        if box is not None:
            for value in box:
                comment += "%7.2f" % (value)
        block += comment
    block += "\n"
    for el, crd in zip(elements, coords):
        block += "%8s " % (el)
        for x in crd:
            block += "%20.10f " % (x)
        block += "\n"
    return block


def data_from_xyzcomment(line):
    """
    Convert XYZ comment line (cell angles are in radians)
    """
    words = line.split()
    if len(words) == 0:
        return {}
    if len(words) < 3:
        return {"Line": line}
    xyzdic = {}
    xyzdic["Name"] = words[0]
    try:
        xyzdic["Step"] = int(words[1])
        xyzdic["Energy"] = float(words[2])
    except ValueError:
        xyzdic["Line"] = line
        return xyzdic
    if len(words) >= 6:
        try:
            box = [float(w) for w in words[3:6]]
        except ValueError:
            return xyzdic
        if len(words) >= 9:
            try:
                angles = [float(w) for w in words[6:9]]
            except ValueError:
                return xyzdic
            angles = [a for a in angles]
            box += angles
        lattice = cellvectors_from_shape(box)
        xyzdic["Lattice"] = lattice
    return xyzdic



""" Deprecated, do not use """

import argparse
import os

from scm.plams import (
    AMSJob,
    CRSJob,
    KFFile,
    Molecule,
    Results,
    Settings,
    Units,
    finish,
    init,
)

# import redox_defaults as defaults


__all__ = ["redox_potential"]


### ==== SETTINGS ==== ###


### ==== DFT SETTINGS ==== ###
def DFT_defaults():
    s = Settings()
    s.input.ams.task = "GeometryOptimization"
    s.input.adf.basis.type = "TZ2P"
    s.input.adf.basis.core = "None"
    s.input.adf.xc.hybrid = "B3LYP"
    s.input.adf.xc.Dispersion = "GRIMME3 BJDAMP"
    s.input.adf.Relativity.Level = "None"
    s.input.adf.NumericalQuality = "Good"
    s.input.adf.Symmetry = "NOSYM"
    s.input.ams.UseSymmetry = "No"
    s.input.adf.Unrestricted = "No"
    s.input.adf.SpinPolarization = 0
    s.input.ams.System.Charge = 0
    return s

    ### ==== DFTB SETTINGS ==== ###


def DFTB_defaults():
    s = Settings()
    s.input.ams.task = "GeometryOptimization"
    s.input.DFTB
    s.input.DFTB.Model = "GFN1-xTB"
    s.input.ams.System.Charge = 0
    return s

    ### ==== FREQ SETTINGS ==== ###


def frequencies_defaults():
    s = Settings()
    s.input.ams.properties.NormalModes = "Yes"
    s.input.ams.Properties.PESPointCharacter = "No"
    s.input.ams.NormalModes.ReScanFreqRange = "-1000 0"
    s.input.ams.PESPointCharacter.NegativeEigenvalueTolerance = -0.001
    return s


def get_settings(
    molecule: Molecule = None,
    task: str = "GeometryOptimization",
    state: str = "neutral",
    phase: str = "vacuum",
    frequencies: bool = True,
    use_dftb: bool = False,
    use_COSMORS: bool = False,
    init_charge: int = 0,
) -> Settings:
    """
    Method that generates settings for jobs based on settings provided
    """

    if use_COSMORS:
        sett = DFT_defaults()

        solvation_block = {
            "surf": "Delley",
            "solv": "name=CRS cav0=0.0 cav1=0.0",
            "charged": "method=Conj corr",
            "c-mat": "Exact",
            "scf": "Var All",
            "radii": {
                "H": 1.30,
                "C": 2.00,
                "N": 1.83,
                "O": 1.72,
                "F": 1.72,
                "Si": 2.48,
                "P": 2.13,
                "S": 2.16,
                "Cl": 2.05,
                "Br": 2.16,
                "I": 2.32,
            },
        }

        sett.input.adf.solvation = solvation_block
    else:
        if use_dftb:
            sett = DFTB_defaults()
        else:
            sett = DFT_defaults()

        # load cosmo solvent
        if phase != "vacuum" and not use_COSMORS:
            sett.input.adf.Solvation.Solv = f"name={phase}"

    sett.input.ams.task = task

    if frequencies:
        sett.soft_update(frequencies_defaults())

    # set the charge
    if state == "oxidation":
        charge = init_charge + 1
    elif state == "reduction":
        charge = init_charge - 1
    else:
        charge = init_charge

    if not use_dftb:
        # if we are doing DFT we have to change the spinpolarization and whether to use unrestricted DFT
        num_electrons = sum(atom.atnum for atom in molecule) - charge
        spinpol = num_electrons % 2
        if spinpol == 0:
            unrestricted = "No"
        else:
            unrestricted = "Yes"
        sett.input.adf.SpinPolarization = spinpol
        sett.input.adf.Unrestricted = unrestricted

    sett.input.ams.System.Charge = charge

    return sett


### ==== UTILITY FUNCTIONS ==== ###


def check_termination_succes(result: Results) -> bool:
    term = result.readrkf("General", "termination status", "ams")
    if term == "NORMAL TERMINATION":
        return True
    elif "NORMAL TERMINATION" in term:
        return "WARNING"
    return


### ==== CALCULATIONS ==== ###


def COSMORS_property(solvent_path: str, solute_path: str, name: str, temperature: float = 298.15) -> float:
    """This method runs a COSMORS property job to obtain the activity coefficient
    which will also calculate G solute which we need to calculate the redox
    potential.
    """

    sett = Settings()
    sett.input.property._h = "ACTIVITYCOEF"
    compounds = [Settings(), Settings()]
    compounds[0]._h = solvent_path
    compounds[1]._h = solute_path
    compounds[0].frac1 = 1
    compounds[1].frac1 = 0

    sett.input.temperature = str(temperature)
    sett.input.compound = compounds

    res = CRSJob(settings=sett, name=name).run().get_results()
    if res:
        # convert the Gibbs energy to hartree (COSMORS gives in kcal/mol)
        return float(Units.convert(res["G solute"][1], "kcal/mol", "hartree"))
    else:
        return False


def calculation_step(
    molecule: Molecule,
    task: str = "GeometryOptimization",
    state: str = "neutral",
    phase: str = "vacuum",
    frequencies: bool = False,
    use_dftb: bool = False,
    use_COSMORS: bool = False,
    name: str = None,
    solvent_path: str = None,
    init_charge: int = 0,
) -> dict:
    """Method used to optimize the geometry of molecule using DFT (by default B3LYP)
    Other settings may be supplied using settings which will be soft-updated using DFT_defaults
    State specifies whether the molecule is neutral, oxidised or reduced
    Phase specifies whether the system is in vacuum or solvated
    if use_dftb, the system will be optimised using DFTB (by default GFN1-xTB) instead of DFT
    """

    if use_COSMORS:
        phase = "solvent"

    settings = get_settings(
        molecule=molecule,
        task=task,
        use_dftb=use_dftb,
        use_COSMORS=use_COSMORS,
        state=state,
        phase=phase,
        frequencies=frequencies,
        init_charge=init_charge,
    )

    # summarize job in one string
    task_abbrev = {"GeometryOptimization": "GO", "SinglePoint": "SP"}[task]
    job_desc = f"{task_abbrev}_{state}_{phase}"
    if use_COSMORS:
        job_desc += "_COSMO-RS"
    if use_dftb:
        job_desc += "_DFTB"

    print(f'\nStarting calculation {name + "_" + job_desc}')
    print(f"\ttask                 = {task}")
    print(f"\tuse_dftb             = {use_dftb}")
    print(f"\tuse_COSMORS          = {use_COSMORS}")
    print(f"\tfrequencies          = {frequencies}")
    print(f"\tstate                = {state}")
    print(f"\tphase                = {phase}")
    print(f"\tcharge               = {settings.input.ams.System.Charge}")
    if not use_dftb:
        print(f"\tunrestricted         = {settings.input.adf.Unrestricted}")
        print(f"\tspin polarization    = {settings.input.adf.SpinPolarization}")

    # run the job
    job = AMSJob(molecule=molecule, settings=settings, name=name + "_" + job_desc)
    res = job.run()

    result_dict = {}
    # pull out results
    if check_termination_succes(res):
        print(f"\tSuccessfull          = {check_termination_succes(res)}")  # True or WARNING
        # set some default values
        bond_energy = None
        gibbs_energy = None

        # If we are doing COSMO calculations then we need to run an additional job to obtain the activity coefficient
        # when calculating the activity coefficient, the G solute is also calculated.
        if use_COSMORS:
            resfile = KFFile(res["adf.rkf"])
            cosmo_data = resfile.read_section("COSMO")
            coskf = KFFile(os.path.join(job.path, "solute.coskf"))
            for k, v in cosmo_data.items():
                coskf.write("COSMO", k, v)
            res.collect()
            bond_energy = res.readrkf("AMSResults", "Energy", "adf")
            solute_path = os.path.join(job.path, "solute.coskf")
            gibbs_energy = COSMORS_property(solvent_path, solute_path, job.name + "_ACTIVITYCOEF")
        # if we dont use COSMO-RS we can just extract the Gibbs and bonding energies from the regular job
        else:
            if use_dftb:
                bond_energy = res.readrkf("AMSResults", "Energy", "dftb")
                if frequencies:
                    gibbs_energy = res.readrkf("Thermodynamics", "Gibbs free Energy", "dftb")
            else:
                bond_energy = res.readrkf("Energy", "Bond Energy", "adf")
                if frequencies:
                    gibbs_energy = res.readrkf("Thermodynamics", "Gibbs free Energy", "adf")

        print("\tResults:")
        if not bond_energy is None:
            result_dict["bond_energy"] = Units.convert(bond_energy, "hartree", "eV")
            print(f'\t\tBond Energy  = {result_dict["bond_energy"]:.4f} eV')
        if not gibbs_energy is None:
            result_dict["gibbs_energy"] = Units.convert(gibbs_energy, "hartree", "eV")
            print(f'\t\tGibbs Energy = {result_dict["gibbs_energy"]:.4f} eV')

        # extract also optimised molecule
        if task == "GeometryOptimization":
            result_dict["geometry"] = res.get_main_molecule()

        # and if the phase is solvent we also need the solvation gibbs energy change
        if phase != "vacuum":
            dG_solvation = res.readrkf("Energy", "Solvation Energy (el)", "adf") + res.readrkf(
                "Energy", "Solvation Energy (cd)", "adf"
            )
            result_dict["dG_solvation"] = Units.convert(dG_solvation, "hartree", "eV")
            print(f'\t\tdG_solvation = {result_dict["dG_solvation"]:.4f} eV')

    else:
        print("\tSuccessfull          = False")

    return result_dict


### ==== MAIN FUNCTION ==== ###


def redox_potential(
    molecule: Molecule,
    mode: str,
    method: str = "screening",
    name: str = None,
    COSMORS_solvent_path: str = None,
    solvent: str = "Dichloromethane",
    init_charge: int = 0,
) -> float:

    assert mode in ["oxidation", "reduction"]
    assert method in [
        "DC",
        "TC-COSMO",
        "TC-COSMO-RS",
        "screening",
    ], 'Argument "method" must be "DC", "TC-COSMO", "TC-COSMO-RS" or "screening"'
    assert os.path.exists(COSMORS_solvent_path), f"Could not find coskf file {COSMORS_solvent_path}"

    # set name
    if name is None:
        name = molecule.properties.name

    print("========================================================================")
    print(f"Starting redox potential calculation for molecule {name}:\n")

    print("\nInitial coordinates:")
    print(molecule)
    print("Settings:")
    print(f"\tName:              {name}")
    print(f"\tMode:              {mode}")
    print(f"\tMethod:            {method}")
    print(f"\tSolvent:           {solvent}")
    print(f"\tInitial Charge:    {init_charge}")

    if mode == "oxidation":
        Gelectron = -0.0375
    elif mode == "reduction":
        Gelectron = 0.0375

    # some settings that every calculation step requires
    general_settings = {
        "name": name,
        "init_charge": init_charge,
    }

    # get on with the actual calculations
    if method == "DC":
        GO_os = calculation_step(molecule, state=mode, phase=solvent, frequencies=True, **general_settings)
        GO_ns = calculation_step(molecule, phase=solvent, frequencies=True, **general_settings)

        redoxpot = GO_os["gibbs_energy"] - GO_ns["gibbs_energy"] + Gelectron

    elif method == "TC-COSMO":
        GO_nv = calculation_step(molecule, frequencies=True, **general_settings)
        SP_nv_ns = calculation_step(GO_nv["geometry"], task="SinglePoint", phase=solvent, **general_settings)
        GO_ns = calculation_step(molecule, phase=solvent, **general_settings)
        SP_nv_nv = calculation_step(GO_ns["geometry"], task="SinglePoint", **general_settings)

        GO_ov = calculation_step(molecule, state=mode, frequencies=True, **general_settings)
        SP_ov_os = calculation_step(
            GO_ov["geometry"], task="SinglePoint", state=mode, phase=solvent, **general_settings
        )
        GO_os = calculation_step(molecule, state=mode, phase=solvent, **general_settings)
        SP_ov_ov = calculation_step(GO_os["geometry"], task="SinglePoint", state=mode, **general_settings)

        redox_part = GO_ov["gibbs_energy"] + SP_ov_os["dG_solvation"] + (SP_ov_ov["bond_energy"] - GO_ov["bond_energy"])
        neutral_part = (
            GO_nv["gibbs_energy"] + SP_nv_ns["dG_solvation"] + (SP_nv_nv["bond_energy"] - GO_nv["bond_energy"])
        )
        redoxpot = redox_part - neutral_part + Gelectron

    elif method == "TC-COSMO-RS":
        assert os.path.exists(COSMORS_solvent_path), f"Solvent database {COSMORS_solvent_path} does not exist"
        GO_nv = calculation_step(molecule, frequencies=True, **general_settings)
        SP_nv_ns = calculation_step(
            GO_nv["geometry"],
            task="SinglePoint",
            use_COSMORS=True,
            solvent_path=COSMORS_solvent_path,
            **general_settings,
        )
        GO_ns = calculation_step(molecule, use_COSMORS=True, solvent_path=COSMORS_solvent_path, **general_settings)
        SP_ns_nv = calculation_step(GO_ns["geometry"], task="SinglePoint", **general_settings)

        GO_ov = calculation_step(molecule, state=mode, frequencies=True, **general_settings)
        SP_ov_os = calculation_step(
            GO_ov["geometry"],
            task="SinglePoint",
            state=mode,
            use_COSMORS=True,
            solvent_path=COSMORS_solvent_path,
            **general_settings,
        )
        GO_os = calculation_step(
            molecule, state=mode, use_COSMORS=True, solvent_path=COSMORS_solvent_path, **general_settings
        )
        SP_os_ov = calculation_step(GO_os["geometry"], task="SinglePoint", state=mode, **general_settings)

        redox_part = GO_ov["gibbs_energy"] + SP_ov_os["dG_solvation"] + (SP_os_ov["bond_energy"] - GO_ov["bond_energy"])
        neutral_part = (
            GO_nv["gibbs_energy"] + SP_nv_ns["dG_solvation"] + (SP_ns_nv["bond_energy"] - GO_nv["bond_energy"])
        )
        redoxpot = redox_part - neutral_part + Gelectron

    elif method == "screening":
        assert os.path.exists(COSMORS_solvent_path), f"Solvent database {COSMORS_solvent_path} does not exist"
        GO_nv = calculation_step(molecule, use_dftb=True, **general_settings)
        COSMO_nv = calculation_step(
            GO_nv["geometry"],
            task="SinglePoint",
            use_COSMORS=True,
            solvent_path=COSMORS_solvent_path,
            **general_settings,
        )

        GO_ov = calculation_step(molecule, state=mode, use_dftb=True, **general_settings)
        COSMO_ov = calculation_step(
            GO_ov["geometry"],
            task="SinglePoint",
            state=mode,
            use_COSMORS=True,
            solvent_path=COSMORS_solvent_path,
            **general_settings,
        )

        redoxpot = COSMO_ov["gibbs_energy"] - COSMO_nv["gibbs_energy"] + Gelectron

    print(f"\nOxidation potential: {redoxpot:.4f} eV")
    return redoxpot


def main(mode):
    job_dir = "./Test"
    solvent = "DMSO"
    mol_files = [
        "./molecules/20410-46-2.xyz",
        "./molecules/109-75-1.xyz",
        "./molecules/167951-80-6.xyz",
        "./molecules/92-52-4.xyz",
    ]
    mol_names = [
        "Phosphorodifluoridate (8CI,9CI)",
        "3-Butenenitrile",
        "3,3,3-Trifluoropropylene carbonate",
        "1,1-Biphenyl",
    ]
    init_charges = [-1, 0, 0, 0]

    if not os.path.exists(job_dir):
        os.makedirs(job_dir)

    COSMORS_solvent_path = os.path.abspath(f"coskf/{solvent}.coskf")

    results = {}
    for file, name, charge in zip(mol_files, mol_names, init_charges):
        results[name] = {}
        for method in ["screening", "TC-COSMO", "TC-COSMO-RS", "DC"]:

            job_name = f"{name}_{method}_{mode}"

            # calculation part
            init(path=job_dir, folder=job_name)

            mol = Molecule(file)
            redoxpot = redox_potential(
                mol,
                mode,
                name=name,
                method=method,
                COSMORS_solvent_path=COSMORS_solvent_path,
                solvent=solvent,
                init_charge=charge,
            )
            results[name][method] = redoxpot

            finish()

    ### PRINT POTENTIALS
    print(f"{mode.capitalize()} Potentials:")
    name_len = max(len("System"), max(len(n) for n in results))
    methods = list(results[list(results.keys())[0]])
    method_lens = [max(9, len(m)) for m in methods]
    print(f'{"System".ljust(name_len)} | {" | ".join([m.ljust(l) for m, l in zip(methods, method_lens)])}')
    for name, res in results.items():
        s = f'{name.ljust(name_len)} | {" | ".join([(str(round(res[m],3)) + " V").rjust(l) for m, l in zip(methods, method_lens)])}'
        print(s)

    print("\nCalculations coomplete!\a")


if __name__ == "__main__":
    ap = argparse.ArgumentParser(description="Redox potential workflow.")
    ap.add_argument("molecule", type=str)
    main("oxidation")



#!/usr/bin/env amspython
from abc import ABC, abstractmethod
from collections import defaultdict, namedtuple
from dataclasses import dataclass
from natsort import natsorted
from pathlib import Path
from typing import Dict, List, Tuple
import numpy as np
import scm.plams as plams
import scm.reactmap
import warnings

__all__ = ["SMILESReaction", "XYZReaction", "MoleculeReaction", "run_ts_workflow"]


@dataclass
class BoostReaction(ABC):
    @abstractmethod
    def get_molecules_dict(self) -> Dict[str, plams.Molecule]:
        pass

    @staticmethod
    def mapping(reactant: plams.Molecule, product: plams.Molecule) -> plams.Molecule:
        """returns a new product with rearranged atom indices"""
        settings = scm.reactmap.Settings(print_progress=True, preprocess_map=True)
        reaction = scm.reactmap.Reaction(
            reactant=scm.reactmap.Molecule(plams_mol=reactant), product=scm.reactmap.Molecule(plams_mol=product)
        )
        if not reaction.is_feasible():
            raise ValueError(f"Unfeasible reaction: {reaction}")
        scm.reactmap.Map(reaction, settings=settings)

        plams.log("SCM Reaction Mapping enabled!")
        plams.log(reaction.mapping)
        for i, x in enumerate(reaction.mapping):
            if i != x:
                plams.log(f"Reactant atom {i+1} -> product atom {x+1}")

        temp_mol = product.get_fragment(reaction.mapping)  # get_fragment uses zero-based indices!
        return temp_mol


def reaction_examples(i: int) -> BoostReaction:
    reactions = {
        0: SMILESReaction(
            reactants_smiles=["C1=CC(=O)C(=O)C=C1", "[OH3+]"],
            # products_smiles = ["C1=C[C]([OH+])C(=O)C=C1", "O"],
            products_smiles=["C1=CC(=[OH+])C(=O)C=C1", "O"],
        ),
        1: SMILESReaction(
            reactants_smiles=["C1=CC(=[OH+])C(=O)C=C1", "O"],
            products_smiles=["C1=CC(O)C(=[OH+])C(O)=C1"],
        ),
        2: SMILESReaction(
            reactants_smiles=["C=C", "O"],
            products_smiles=["CCO"],
        ),
        3: SMILESReaction(
            reactants_smiles=["C1C=CC=C1", "C1=CC(=O)OC1=O"],
            products_smiles=["[H][C@@]12C3CC(C=C3)[C@]1([H])C(=O)OC2=O"],
        ),
    }
    return reactions[i]


def uff_settings() -> plams.Settings:
    """returns UFF engine settings"""
    s = plams.Settings()
    s.input.ForceField.Type = "UFF"
    s.runscript.nproc = 1
    return s


def run_ts_workflow(
    reaction: BoostReaction,
    engine_settings: plams.Settings,
    reaction_boost_equilibration_nsteps: int = 1000,
    reaction_boost_nsteps: int = 500,
    reaction_boost_strength: float = 0.7,
    reaction_boost_equilibration_uff: bool = True,
    reaction_boost_equilibration_strength: float = 0.05,
    do_ts_search: bool = True,
    do_irc: bool = True,
    run_equilibration: bool = True,
    temperature: float = 200,
):
    """Runs all the steps"""

    warnings.filterwarnings("ignore", "biadjacency_matrix will return a scipy.sparse array instead of a matrix")

    if reaction_boost_equilibration_uff:
        equilibration_engine_settings = uff_settings()
    else:
        equilibration_engine_settings = engine_settings.copy()

    molecules_dict = reaction.get_molecules_dict()
    if "final" in molecules_dict and False:
        molecules_dict["final"] = plams.preoptimize(molecules_dict["final"], settings=engine_settings, maxiterations=30)

    boost_job = run_reaction(
        molecules_dict,
        engine_settings=engine_settings,
        run_equilibration=run_equilibration,
        equilibration_nsteps=reaction_boost_equilibration_nsteps,
        equilibration_engine_settings=equilibration_engine_settings,
        equilibration_strength=reaction_boost_equilibration_strength,
        preoptimize_after_equilibration=True,
        nsteps=reaction_boost_nsteps,
        strength=reaction_boost_strength,
        temperature=temperature,
    )

    md_forward_barrier, md_backward_barrier = boost_job.get_barriers()
    plams.log(f"Forward MD: {md_forward_barrier*27.211:.3f}")
    plams.log(f"Backward MD: {md_backward_barrier*27.211:.3f}")

    ts_job = run_ts(
        boost_job,
        engine_settings=engine_settings,
    )

    if not ts_job.ok():
        plams.log("No TS found, perhaps barrierless?!")
        return

    irc_job = get_irc_job(
        ts_job,
        engine_settings=engine_settings,
    )
    irc_job.run()

    if not irc_job.ok():
        plams.log("IRC job failed!")
        return

    irc_results = irc_job.results.get_irc_results()
    irc_forward_barrier = irc_results["LeftBarrier"]
    irc_backward_barrier = irc_results["RightBarrier"]

    plams.log(f"Forward IRC: {irc_backward_barrier*27.211:.3f}")
    plams.log(f"Backward IRC: {irc_forward_barrier*27.211:.3f}")


@dataclass
class SMILESReaction(BoostReaction):
    reactants_smiles: List[str]
    products_smiles: List[str]
    name: str = "reaction"
    fix_stoichiometry: bool = True

    def get_molecules(self) -> Tuple[plams.Molecule, plams.Molecule]:
        if self.fix_stoichiometry:
            self.add_molecules_if_needed()

        reactants = self.get_combined(self.reactants_smiles)
        products = self.get_combined(self.products_smiles)
        products = self.mapping(reactants, products)
        return reactants, products

    def _fix_single_stoichoimetry(self, stoich: Dict[str, int]) -> bool:
        """Returns True if the stoichiometry was modified"""
        A = namedtuple("A", ["stoich", "smiles"])
        fixes = [
            A(stoich={"H": 2, "O": 1}, smiles="O"),
            A(stoich={"H": 1, "Cl": 1}, smiles="Cl"),
            A(stoich={"H": 1, "F": 1}, smiles="F"),
            A(stoich={"H": 1, "Br": 1}, smiles="Br"),
            A(stoich={"H": 2, "S": 1}, smiles="S"),
        ]

        for fix in fixes:
            if all(stoich[x] <= -y for x, y in fix.stoich.items()):
                self.products_smiles.append(fix.smiles)
                for k, v in fix.stoich.items():
                    stoich[k] += v
                return True
            if all(stoich[x] >= y for x, y in fix.stoich.items()):
                self.reactants_smiles.append(fix.smiles)
                for k, v in fix.stoich.items():
                    stoich[k] -= v
                return True
        return False

    def add_molecules_if_needed(self):
        """modifies self.reactants_smiles and self.products_smiles"""
        rmols = [plams.from_smiles(x) for x in self.reactants_smiles]
        pmols = [plams.from_smiles(y) for y in self.products_smiles]
        stoich = defaultdict(lambda: 0)
        for mol in rmols:
            for k, v in mol.get_formula(as_dict=True).items():
                stoich[k] -= v
        for mol in pmols:
            for k, v in mol.get_formula(as_dict=True).items():
                stoich[k] += v

        while self._fix_single_stoichoimetry(stoich):
            pass

    def get_molecules_dict(self) -> Dict[str, plams.Molecule]:
        r, p = self.get_molecules()
        ret = {"": r, "final": p}
        return ret

    @classmethod
    def get_combined(cls, smiles_list: List[str], margin: float = 1.0) -> plams.Molecule:
        ret = plams.Molecule()
        molecules = [plams.from_smiles(x) for x in smiles_list]
        tot_charge = sum(mol.properties.get("charge", 0) for mol in molecules)
        molecules = plams.preoptimize(molecules)
        for m in molecules:
            ret.add_molecule(m, margin=margin)
        ret.properties.charge = tot_charge
        return ret


@dataclass
class MoleculeReaction(BoostReaction):
    reactant: plams.Molecule
    product: plams.Molecule
    do_mapping: bool = False

    def get_molecules_dict(self) -> Dict[str, plams.Molecule]:
        """Returns a molecule dictionary"""
        if self.do_mapping:
            product = self.mapping(self.reactant, self.product)
        else:
            product = self.product
        ret = {"": self.reactant, "final": self.mapping(self.reactant, product)}
        return ret


@dataclass
class XYZReaction(BoostReaction):
    folder: Path  # a folder with .xyz files
    charge: float = 0

    def get_molecules_dict(self) -> Dict[str, plams.Molecule]:
        """
        read molecules from .xyz files in a directory ``self.folder``.
        Returns a dictionary suitable for AMSJob
        """
        # read all xyz files, dictionary key: Molecule
        molecules = plams.read_molecules(str(self.folder.resolve()))

        sorted_keys = list(natsorted(molecules.keys()))
        molecules[""] = molecules.pop(sorted_keys[0])  # the first molecule must have key ''
        sorted_keys.pop(0)

        if self.charge != 0:
            for mol in molecules.values():
                mol.properties.charge = self.charge

        # molecules[""].hydrogen_to_deuterium()  # slow down hydrogen motions
        # molecules[''].lattice = [[9, 0, 0], [0, 9, 0], [0, 0, 9]]  # optional, add lattice

        return molecules


class AMSReactionBoostJob(plams.AMSMDJob):
    def __init__(
        self,
        molecule,
        settings: plams.Settings = None,
        nsteps_per: int = None,
        nsteps: int = None,
        strength: float = 1.0,
        initial_fraction: float = 0.05,
        post_reaction_relaxation_steps: int = 0,
        **kwargs,
    ):
        if isinstance(molecule, plams.Molecule):
            molecule = {"": molecule}
        assert isinstance(molecule, dict)
        assert len(molecule) >= 2

        n_transitions = len(molecule) - 1
        if nsteps_per is None and nsteps is not None:
            nsteps_per = round(nsteps / n_transitions)
        elif nsteps is None and nsteps_per is not None:
            nsteps = nsteps_per * n_transitions + post_reaction_relaxation_steps

        settings = settings or dict()

        super().__init__(molecule=molecule, settings=settings, nsteps=nsteps, **kwargs)
        self.nsteps_per = nsteps_per
        self.strength = strength
        self.post_reaction_relaxation_steps = post_reaction_relaxation_steps
        self.n_transitions = n_transitions
        self.initial_fraction = initial_fraction

        self.settings += self.get_boost_settings(
            what="pair", strength=self.strength, initial_fraction=self.initial_fraction
        )

        self.sorted_keys = [x for x in molecule if x != ""]
        self.settings.input.ams.MolecularDynamics.ReactionBoost.TargetSystem = self.sorted_keys
        self.settings.input.ams.MolecularDynamics.ReactionBoost.NSteps = self.nsteps_per

    @classmethod
    def get_boost_settings(cls, what: str, strength: float = 1.0, initial_fraction: float = 0.05) -> plams.Settings:
        """Returns the bulk of the MolecularDynamics%ReactionBoost settings"""
        s = plams.Settings()
        s.input.ams.Log.Info = "ReactionBoost"
        if what.lower() == "rmsd":
            s.input.ams.MolecularDynamics.ReactionBoost.Type = "RMSD"
            s.input.ams.MolecularDynamics.ReactionBoost.RMSDRestraint.Type = "GaussianWell"
            # s.input.ams.MolecularDynamics.ReactionBoost.Region = "solute"
            s.input.ams.MolecularDynamics.ReactionBoost.Change = "TargetCoordinate"
            # s.input.ams.MolecularDynamics.ReactionBoost.InitialFraction = 0.01
            # s.input.ams.MolecularDynamics.ReactionBoost.RMSDRestraint.Erf.MaxForce = 0.1
        else:
            s.input.ams.MolecularDynamics.ReactionBoost.Type = "Pair"
            rb = s.input.ams.MolecularDynamics.ReactionBoost
            rb.Change = "LogForce"
            rb.InitialFraction = initial_fraction  # at the first time step

            rb.BondBreakingRestraints.Type = "Erf"  # "None" to disable
            # rb.BondBreakingRestraints.Type = "None"  # "None" to disable
            rb.BondBreakingRestraints.Erf.MaxForce = 0.05 * strength
            # rb.BondBreakingRestraints.Erf.ForceConstant = 0.1

            rb.BondMakingRestraints.Type = "Erf"  # "None" to disable
            rb.BondMakingRestraints.Erf.MaxForce = 0.2 * strength
            # rb.BondMakingRestraints.Erf.ForceConstant = 0.1

            rb.BondedRestraints.Type = "Harmonic"  # "None" to disable
            rb.BondedRestraints.Harmonic.ForceConstant = 0.001 * strength

            rb.NonBondedRestraints.Type = "Exponential"  # "None" to disable
            rb.NonBondedRestraints.Exponential.Epsilon = 1e-4 * strength

        return s

    def get_approximate_ts_index(self) -> int:
        """Returns 1-based index for frame with highest engine energy"""
        engine_energy_hartree = self.results.get_history_property("EngineEnergy", "History")
        index = np.argmax(engine_energy_hartree) + 1

        return index

    def get_barriers(self, index: int = None) -> Tuple[float, float]:
        """Returns the forward and backward potential energy barriers in hartree relative to the frame with index=index"""
        if index is None:
            index = self.get_approximate_ts_index()

        engine_energy_hartree = self.results.get_history_property("EngineEnergy")
        md_forward_barrier = np.max(engine_energy_hartree) - np.min(engine_energy_hartree[:index])
        md_backward_barrier = np.max(engine_energy_hartree) - np.min(engine_energy_hartree[index - 1 :])

        return md_forward_barrier, md_backward_barrier

    def get_reaction_coordinates(
        self,
    ) -> Tuple[
        List[Tuple[int, int, float]],  # bond making
        List[Tuple[int, int, float]],  # bond breaking
        List[Tuple[int, int, float]],  # bonded
    ]:
        """
        Reads the stdout log info and finds out which restraints were added

        NOTE: This assumes that n_transitions = 1!
        """

        bond_making_lines = self.results.grep_output("bond-making")
        bond_breaking_lines = self.results.grep_output("bond-breaking")
        bonded_lines = self.results.grep_output("bonded")

        bond_making_tuples = []
        bond_breaking_tuples = []
        bonded_tuples = []

        for line in bond_making_lines:
            splitline = line.split()
            target_value = float(splitline[5])  # bohr
            if target_value > 5:
                continue
            bond_making_tuples.append((int(splitline[2]), int(splitline[3]), 1.0))

        for line in bond_breaking_lines:
            splitline = line.split()  # missing space so use other indices
            initial_value = float(splitline[3])  # bohr
            if initial_value > 5:
                continue
            bond_breaking_tuples.append((int(splitline[1]), int(splitline[2]), -1.0))

        for line in bonded_lines:
            splitline = line.split()
            mean = 0.5 * (float(splitline[4]) + float(splitline[5])) * 0.529
            bonded_tuples.append((int(splitline[2]), int(splitline[3]), mean))

        return bond_making_tuples, bond_breaking_tuples, bonded_tuples

    def get_rc_atoms(self) -> List[int]:
        """
        Returns atom indices that are involved in bond-making or bond-breaking restraints
        """

        bm, bb, _ = self.get_reaction_coordinates()
        ret = set()
        for b in bm + bb:
            ret.add(b[0])
            ret.add(b[1])
        return list(ret)

    def get_reaction_coordinate_lines(self) -> Tuple[List[str], List[str], List[str]]:
        """
        Returns 3-tuple list of str: bond-making, bond-breaking, bonded
        Can be used e.g. to set
        s.input.ams.TransitionStateSearch.ReactionCoordinate.Distance = (
            bond_making_lines + bond_breaking_lines
        )
        s.input.ams.Constraints.Distance = bonded_lines
        """

        def rc_tuple_to_str_list(rc_tuple):
            if rc_tuple is None:
                return []
            ret = []
            for t in rc_tuple:
                ret.append(" ".join(str(x) for x in t))

            return ret

        bond_making, bond_breaking, bonded = self.get_reaction_coordinates()

        bond_making_lines = rc_tuple_to_str_list(bond_making)
        bond_breaking_lines = rc_tuple_to_str_list(bond_breaking)
        bonded_lines = rc_tuple_to_str_list(bonded)

        return bond_making_lines, bond_breaking_lines, bonded_lines


def run_reaction(
    molecules_dict: Dict[str, plams.Molecule],
    engine_settings: plams.Settings,
    run_equilibration: bool = True,
    equilibration_nsteps: int = 500,
    equilibration_strength: float = 0.05,
    equilibration_engine_settings: plams.Settings = None,
    nsteps: int = 1000,
    strength: float = 1.0,
    preoptimize_after_equilibration: bool = True,
    temperature: float = 200,
) -> AMSReactionBoostJob:
    """Runs equilibration and produciton boost MD. Returns the production job"""

    for i, at in enumerate(molecules_dict[""], 1):
        if (
            "Constraints" in engine_settings.input.ams
            and "Atom" in engine_settings.input.ams.Constraints
            and i not in engine_settings.input.ams.Constraints.Atom
        ):
            plams.AMSJob._add_region(at, "thermostatted")
        else:
            plams.AMSJob._add_region(at, "frozen")

    if run_equilibration:
        equilibration_engine_settings = equilibration_engine_settings or engine_settings.copy()
        eq_job = AMSReactionBoostJob(
            settings=equilibration_engine_settings,
            molecule=molecules_dict,
            name="eq",
            nsteps=equilibration_nsteps,
            strength=equilibration_strength,
            temperature=300,
            samplingfreq=20,
            tau=5,  # very short time constant (fs) to cool the system down
            timestep=1.0,  # fs
            thermostat="NHC",
            thermostat_region="thermostatted",
        )
        eq_job.run()

        # update the initial molecule
        molecules_dict[""] = eq_job.results.get_main_molecule()

        if preoptimize_after_equilibration:
            rc_atoms = eq_job.get_rc_atoms()
            preoptimization_job = run_preoptimization_job_fixing_active_atoms(
                engine_settings=engine_settings,
                molecule=molecules_dict[""],
                rc_atoms=rc_atoms,
                name="preoptimization_prod",
            )

            molecules_dict[""] = preoptimization_job.results.get_main_molecule()

    prod_job = AMSReactionBoostJob(
        settings=engine_settings,
        molecule=molecules_dict,
        name="prod",
        nsteps=nsteps,
        strength=strength,
        post_reaction_relaxation_steps=20,
        temperature=temperature,
        samplingfreq=10,
        tau=3.0,  # very short time constant (fs) to cool the system down
        timestep=0.5,
        thermostat="NHC",
        writeenginegradients=True,
        thermostat_region="thermostatted",
    )

    prod_job.run()

    return prod_job


def run_preoptimization_job_fixing_active_atoms(
    rc_atoms: List[int], molecule: plams.Molecule, engine_settings: plams.Settings, name: str = "preoptimization"
) -> plams.AMSJob:

    for at_i in rc_atoms:
        plams.AMSJob._add_region(molecule[at_i], "bondchange")

    s = plams.Settings()
    s += engine_settings
    s.input.ams.GeometryOptimization.Method = "Quasi-Newton"
    s.input.ams.GeometryOptimization.CoordinateType = "Auto"
    s.input.ams.GeometryOptimization.MaxIterations = 500
    s.input.ams.GeometryOptimization.PretendConverged = "Yes"

    s.input.ams.Task = "GeometryOptimization"
    s.input.ams.GeometryOptimization.Convergence.Quality = "Basic"
    s.input.ams.Constraints.Atom = rc_atoms
    preoptimization_job = plams.AMSJob(
        settings=s,
        molecule=molecule,
        name=name,
    )
    preoptimization_job.run()

    return preoptimization_job


def run_ts(
    boost_job: AMSReactionBoostJob,
    engine_settings: plams.Settings,
    coordinate_type: str = "Auto",
    preoptimization: bool = True,
) -> plams.AMSJob:
    """Returns a transition state search job based on a preiovus AMSReactionBoostJob"""

    index = boost_job.get_approximate_ts_index()
    approximate_ts_molecule = boost_job.results.get_history_molecule(index)

    if isinstance(boost_job.molecule, dict) and len(boost_job.molecule) == 2:
        bond_making_lines, bond_breaking_lines, bonded_lines = boost_job.get_reaction_coordinate_lines()
        rc_atoms = boost_job.get_rc_atoms()
        bonded_lines = []
    else:
        bond_making_lines, bond_breaking_lines, bonded_lines = [], [], []
        rc_atoms = []

    use_reaction_coordinates = False

    s = plams.Settings()
    s += engine_settings
    s.input.ams.GeometryOptimization.Method = "Quasi-Newton"
    s.input.ams.GeometryOptimization.CoordinateType = coordinate_type
    s.input.ams.GeometryOptimization.MaxIterations = 500
    s.input.ams.GeometryOptimization.PretendConverged = "Yes"

    if preoptimization and rc_atoms:
        preoptimization_job = run_preoptimization_job_fixing_active_atoms(
            molecule=approximate_ts_molecule,
            name="preoptimization_ts",
            engine_settings=engine_settings,
            rc_atoms=rc_atoms,
        )
        approximate_ts_molecule = preoptimization_job.results.get_main_molecule()

    s.input.ams.Task = "TransitionStateSearch"
    s.input.ams.GeometryOptimization.InitialHessian.Type = "Calculate"
    s.input.ams.GeometryOptimization.MaxIterations = 1000
    # s.input.ams.GeometryOptimization.Convergence.Gradients = 1e-4   # hartree/angstrom
    # s.input.ams.GeometryOptimization.Convergence.Energy = 1e-5   # hartree
    # s.input.ams.GeometryOptimization.Convergence.Quality = "Basic"

    if bond_making_lines or bond_breaking_lines and use_reaction_coordinates:
        s.input.ams.TransitionStateSearch.ReactionCoordinate.Distance = bond_making_lines + bond_breaking_lines

    if bonded_lines:
        s.input.ams.Constraints.Distance = bonded_lines

    s.input.ams.Properties.PESPointCharacter = "Yes"
    job = plams.AMSJob(settings=s, molecule=approximate_ts_molecule, name="ts")

    job.run()
    return job


def get_irc_job(ts_job: plams.AMSJob, engine_settings: plams.Settings) -> plams.AMSJob:
    s = plams.Settings()
    s += engine_settings
    s.input.ams.Task = "IRC"
    s.input.ams.IRC.MinEnergyProfile = "Yes"
    s.input.ams.IRC.Direction = "Both"
    s.input.ams.GeometryOptimization.PretendConverged = "Yes"
    s.input.ams.GeometryOptimization.Convergence.Quality = "Basic"

    job = plams.AMSJob(settings=s, molecule=ts_job.results.get_main_molecule(), name="irc")

    return job



#!/usr/bin/env python

import array
import os
import struct

from scm.plams.core.errors import PlamsError
import numpy
from scm.plams.mol.molecule import Molecule
from scm.plams.trajectories.trajectoryfile import TrajectoryFile

__all__ = ["DCDTrajectoryFile"]


class DCDTrajectoryFile(TrajectoryFile):
    """
    Class representing a DCD file containing a molecular trajectory

    An instance of this class has the following attributes:

    *   ``file_object`` -- A Python :py:class:`file` object, referring to the actual DCD file
    *   ``position``    -- The frame to which the cursor is currently pointing in the DCD file
    *   ``mode``        -- Designates whether the file is in read or write mode ('rb' or 'wb')
    *   ``ntap``        -- The number of atoms in the molecular system (needs to be constant throughout)

    An |DCDTrajectoryFile| object behaves very similar to a regular file object.
    It has read and write methods (:meth:`read_next` and :meth:`write_next`)
    that read and write from/to the position of the cursor in the ``file_object`` attribute.
    If the file is in read mode, an additional method :meth:`read_frame` can be used that moves
    the cursor to any frame in the file and reads from there.
    The amount of information stored in memory is kept to a minimum, as only information from the current frame
    is ever stored.

    Reading and writing to and from the files can be done as follows::

        >>> from scm.plams import DCDTrajectoryFile

        >>> dcd = DCDTrajectoryFile('old.dcd')
        >>> mol = dcd.get_plamsmol()

        >>> dcdout = DCDTrajectoryFile('new.dcd',mode='wb',ntap=dcd.ntap)

        >>> for i in range(dcd.get_length()) :
        >>>     crd,cell = dcd.read_frame(i,molecule=mol)
        >>>     dcdout.write_next(molecule=mol)

    The above script reads information from the DCD file old.dcd into the |Molecule| object ``mol``
    in a step-by-step manner.
    The |Molecule| object is then passed to the :meth:`write_next` method of the new |DCDTrajectoryFile|
    object corresponding to the new dcd file ``new.dcd``.

    The exact same result can also be achieved by iterating over the instance as a callable

        >>> dcd = DCDTrajectoryFile('old.dcd')
        >>> mol = dcd.get_plamsmol()

        >>> dcdout = DCDTrajectoryFile('new.dcd',mode='wb',ntap=dcd.ntap)

        >>> for crd,cell in dcd(mol) :
        >>>     dcdout.write_next(molecule=mol)

    This procedure requires all coordinate information to be passed to and from the |Molecule| object
    for each frame, which can be time-consuming.
    It is therefore also possible to bypass the |Molecule| object when reading through the frames::

        >>> dcd = DCDTrajectoryFile('old.dcd')

        >>> dcdout = DCDTrajectoryFile('new.dcd',mode='wb',ntap=dcd.ntap)

        >>> for crd,cell in dcd :
        >>>     dcdout.write_next(coords=crd,cell=cell)
    """

    def __init__(self, filename=None, mode="rb", fileobject=None, ntap=None):
        """
        Initiates a DCDTrajectoryFile object

        * ``filename``   -- The path to the DCD file
        * ``mode``       -- The mode in which to open the DCD file ('rb' or 'wb')
        * ``fileobject`` -- Optionally, a file object can be passed instead (filename needs to be set to None)
        * ``ntap``       -- If the file is in write mode, the number of atoms needs to be passed here
        """
        TrajectoryFile.__init__(self, filename, mode, fileobject, ntap)

        # DCD specific attributes
        self.celldata = False  # Whether the cell data is in the file (only set in read-mode)
        self.byte_order = "@"  # native endian
        self.timesteps = (0, 0, 0)  # (nsteps,startstep,steps_between_saves)
        self.delta = 2.0  # The timestep (if stored)
        self.namnf = 0  # The number of free atoms (if stored)
        self.stepsize = 0  # The number of bytes used to store a single conformation
        self.intsize = struct.calcsize("i")  # The number of bytes in an integer
        self.floatsize = struct.calcsize("f")  # The number of bytes in a float

        # Skip to the trajectory part of the file
        if self.mode == "rb":
            self._read_header()
        elif self.mode == "ab":
            self._move_cursor_to_append_pos()
        # elif self.mode == 'wb' :
        #        self._write_header()

    def set_byteorder(self, byteorder):
        """
        Specifies wether the file to be read from or written to is little endian or big endian

        * ``byteorder`` -- Can be either '@', '<', or '>', denoting native endian, little endian, or big endian
        """
        self.byte_order = byteorder

    def _read_variable(self, size):
        """
        Reads a variable from a binary file

        :note::
                Assumes that the variable has been written on a machine
                with the same bitsizes for the data
        """
        if size == "i":
            varsize = self.intsize
        if size == "f":
            varsize = self.floatsize
        else:
            varsize = struct.calcsize(size)
        var_char = self.file_object.read(varsize)
        fmt = self.byte_order + size
        var = struct.unpack(fmt, var_char)[0]

        return var

    def _write_variable(self, var, size):
        """
        Writes a variable to a binary file
        """
        var_char = struct.pack(size, var)
        self.file_object.write(var_char)

    def _read_header(self):
        """
        Reads in the header information in the binary DCD file
        """
        # This should be an int32
        input_integer = self._read_variable("i")
        if input_integer != 84:
            print(input_integer)
            print("BAD DCD FORMAT")
            raise PlamsError

        # Read CORD string
        car4 = self.file_object.read(4).decode()
        if car4 != "CORD":
            print("BAD DCD FORMAT")
            raise PlamsError

        # Here we read the data of the timesteps
        # NSET: The number of sets of coordinates
        # ISTART: The starting timestep
        # NSAVC: The number of timesteps between dcd saves
        timesteps = array.array("i")
        timesteps.fromfile(self.file_object, 3)
        self.timesteps = tuple(timesteps)
        # print 'timesteps: ',self.timesteps

        # 5 blanc integers
        dummies = array.array("i")
        dummies.fromfile(self.file_object, 5)

        # NAMNF: The number of free atoms
        self.namnf = self._read_variable("i")

        # DELTA: The timestep
        self._read_variable("f")
        # 10 blanc integers
        dummies = array.array("i")
        dummies.fromfile(self.file_object, 10)

        # The end size of the first block
        input_integer = self._read_variable("i")
        if input_integer != 84:
            print("BAD DCD FORMAT")
            raise PlamsError

        ############################################################

        # The size of the next block
        input_integer = self._read_variable("i")

        if (input_integer - 4) % 80 == 0:
            # NTITLE: The number of 80 char title strings
            ntitle = self._read_variable("i")
            for _ in range(ntitle):
                self.file_object.read(80).decode()
            # The end size of this block
            input_integer = self._read_variable("i")
        else:
            print("BAD DCD FORMAT")
            raise PlamsError

        ############################################################

        # This should be an int32
        input_integer = self._read_variable("i")
        if input_integer != 4:
            print("BAD DCD FORMAT")
            raise PlamsError

        # Read in the number of atoms
        self.ntap = self._read_variable("i")
        # print 'ntap: ',self.ntap
        if self.coords.shape == (0, 3):
            self.coords = numpy.zeros((self.ntap, 3))

        # This should be an int32
        input_integer = self._read_variable("i")
        if input_integer != 4:
            print("BAD DCD FORMAT")
            raise PlamsError

        if self.namnf != 0:
            # This should be an int32
            input_integer = self._read_variable("i")
            if input_integer != (self.ntap - self.namnf) * 4:
                print("BAD DCD FORMAT")
                raise PlamsError

            # This should be an int32
            dummies = array.array("i")
            dummies.fromfile(self.file_object, (self.ntap - self.namnf))
            # print 'list',list(dummies)

            # This should be an int32
            input_integer = self._read_variable("i")
            if input_integer != (self.ntap - self.namnf) * 4:
                print("BAD DCD FORMAT")
                raise PlamsError

        self.stepsize += 3 * self.ntap * self.floatsize
        self.stepsize += 5 * self.intsize

        if len(self.elements) != self.ntap:
            self.elements = ["H"] * self.ntap

    def read_next(self, molecule=None, read=True):
        """
        Reads coordinates and lattice info from the current position of the cursor and returns it

        * ``molecule`` -- |Molecule| object in which the new data needs to be stored
        * ``read``     -- If set to False the cursor will move to the next frame without reading
        """
        # Check for end of file
        # This should be an int32
        input_integer_char = self.file_object.read(self.intsize)
        if len(input_integer_char) > 0:
            if self.firsttime:
                self.celldata = False
                input_integer = struct.unpack(self.byte_order + "i", input_integer_char)[0]
                if input_integer == 48:
                    self.celldata = True
        else:
            return None, None

        if not read and not self.firsttime:
            return self._move_cursor_without_reading()

        # Read cell data
        cell = numpy.zeros((3, 3))
        if self.celldata:
            cell_array = array.array("d")
            try:
                cell_array.fromfile(self.file_object, 6)
                if self.byte_order != "@":
                    cell_array.byteswap()
            except EOFError:
                return None, None
            cell[numpy.tril_indices(3)] = cell_array
            self.file_object.read(self.floatsize)

            # Coords
            if self.firsttime:
                input_integer = self._read_variable("i")
            else:
                self.file_object.read(self.intsize)

        if self.firsttime:
            if input_integer != 4 * self.ntap:
                print("BAD DCD FORMAT")
                raise PlamsError

        xcoords = array.array("f")
        try:
            xcoords.fromfile(self.file_object, self.ntap)
            # xcoords = numpy.fromfile(self.file,dtype='f',count=self.ntap)
            if self.byte_order != "@":
                xcoords.byteswap()
        except EOFError:
            return None, None

        for i in range(2):
            if self.firsttime:
                input_integer = self._read_variable("i")
                if input_integer != 4 * self.ntap:
                    print("BAD DCD FORMAT")
                    raise PlamsError
            else:
                self.file_object.read(self.intsize)

        ycoords = array.array("f")
        try:
            ycoords.fromfile(self.file_object, self.ntap)
            if self.byte_order != "@":
                ycoords.byteswap()
        except EOFError:
            return None, None

        for i in range(2):
            if self.firsttime:
                input_integer = self._read_variable("i")
                if input_integer != 4 * self.ntap:
                    print("BAD DCD FORMAT")
                    raise PlamsError
            else:
                self.file_object.read(self.intsize)

        zcoords = array.array("f")
        try:
            zcoords.fromfile(self.file_object, self.ntap)
            if self.byte_order != "@":
                zcoords.byteswap()
        except EOFError:
            return None, None

        if self.firsttime:
            input_integer = self._read_variable("i")
            if input_integer != 4 * self.ntap:
                print("BAD DCD FORMAT")
                raise PlamsError
        else:
            self.file_object.read(self.intsize)

        # Check that self.coords has not been changed
        if not self.coords.shape == (self.ntap, 3):
            raise PlamsError("The coordinate array has been changed outside the class")
        self.coords[:, 0] = xcoords
        self.coords[:, 1] = ycoords
        self.coords[:, 2] = zcoords

        if self.firsttime:
            if self.celldata:
                self.stepsize += 6 * struct.calcsize("d")
                self.stepsize += self.floatsize
                self.stepsize += self.intsize
            self.firsttime = False

        if isinstance(molecule, Molecule):
            self._set_plamsmol(self.coords, cell, molecule)

        self.position += 1

        return self.coords, cell

    def _is_endoffile(self):
        """
        If the end of file is reached, return coords and cell as None
        """
        test = self.file_object.read(self.stepsize)
        return len(test) < self.stepsize

    def _write_header(self, ntap=0):
        """
        Write the header of the DCD file
        """
        self.ntap = ntap
        self.coords = numpy.zeros((self.ntap, 3))

        # This should be an int32
        self._write_variable(84, "i")

        # Read CORD string
        self.file_object.write("CORD".encode())

        # Here we read the data of the timesteps
        # NSET: The number of sets of coordinates
        # ISTART: The starting timestep
        # NSAVC: The number of timesteps between dcd saves
        timesteps = array.array("i", [0, 1, 1])
        self.file_object.write(timesteps)

        # 5 blanc integers
        dummies = array.array("i", [0, 0, 0, 0, 0])
        self.file_object.write(dummies)

        # NAMNF: The number of free atoms
        self._write_variable(0, "i")

        # DELTA: The timestep
        self._write_variable(self.delta, "f")

        # 10 blanc integers
        dummies = array.array("i")
        dummies.append(1)
        for i in range(8):
            dummies.append(0)
        dummies.append(24)
        self.file_object.write(dummies)

        # The end size of the first block
        self._write_variable(84, "i")

        ############################################################

        # The size of the next block
        self._write_variable(84, "i")

        ntitle = 1
        self._write_variable(ntitle, "i")
        title = "REMARK Created with py_md"
        nlet = len(title)
        for i in range(80 - nlet):
            title += " "
        self.file_object.write(title.encode())

        self._write_variable(84, "i")

        ############################################################

        # This should be an int32
        self._write_variable(4, "i")

        # Read in the number of atoms
        self._write_variable(self.ntap, "i")

        # This should be an int32
        self._write_variable(4, "i")

    def write_next(self, coords=None, molecule=None, cell=[0.0, 0.0, 0.0], conect=None):
        """
        Write frame to next position in trajectory file

        * ``coords``   -- A list or numpy array of (``ntap``,3) containing the system coordinates
        * ``molecule`` -- A molecule object to read the molecular data from
        * ``cell``     -- A set of lattice vectors or cell diameters
        * ``conect``   -- A dictionary containing connectivity info (not used)

        .. note::

                Either ``coords`` or ``molecule`` are mandatory arguments
        """
        if isinstance(molecule, Molecule):
            coords, cell = self._read_plamsmol(molecule)[:2]
        cell = self._convert_cell(cell)

        # If this is the first step, write the header
        if self.position == 0:
            self._write_header(len(coords))

        if self.ntap != len(coords):
            print("BAD coordinate  list")
            return

        # Invert the coords list [xcoords,ycoords,zcoords]
        # coords = numpy.array(coords).transpose()

        # This should be an int32
        self._write_variable(48, "i")

        cell = array.array("d", cell)
        self.file_object.write(cell)
        self._write_variable(0, "f")

        # Coords
        self._write_variable(4 * self.ntap, "i")

        xcoords = array.array("f", coords[:, 0])
        self.file_object.write(xcoords)

        for i in range(2):
            self._write_variable(4 * self.ntap, "i")

        ycoords = array.array("f", coords[:, 1])
        self.file_object.write(ycoords)

        for i in range(2):
            self._write_variable(4 * self.ntap, "i")

        zcoords = array.array("f", coords[:, 2])
        self.file_object.write(zcoords)

        self._write_variable(4 * self.ntap, "i")

        self.position += 1

    def _convert_cell(self, cell):
        """
        Check format of cell, and convert to required lower triangle
        """
        if cell is None:
            cell = numpy.zeros((3, 3))[numpy.tril_indices(3)]
        elif len(cell) == 3 and isinstance(cell[0], float) or isinstance(cell[0], numpy.float64):
            cell = [cell[0], 0.0, cell[1], 0.0, 0.0, cell[2]]
        elif len(cell) < 3:
            newcell = numpy.zeros((3, 3))
            newcell[: len(cell)] = cell
            cell = newcell[numpy.tril_indices(3)]
        else:
            cell = numpy.array(cell)
            cell = cell[numpy.tril_indices(3)]
        return cell

    def _rewind_to_first_frame(self):
        """
        Rewind the file to the first frame
        """
        self.file_object.seek(0)
        self.firsttime = True
        self.position = 0
        self.stepsize = 0
        self._read_header()

    def _rewind_n_frames(self, nframes):
        """
        Rewind the file by nframes frames
        """
        whence = os.SEEK_CUR
        stepsize = self.stepsize + self.intsize
        self.file_object.seek(-nframes * stepsize, whence)
        self.position = self.position - nframes



#!/usr/bin/env plams
# Add new results extraction method
@add_to_class(ADFFragmentResults)
def get_transfer_integrals(self):
    return self.job.full.results.read_rkf_section("TransferIntegrals", file="adf")


# Common settings for all 3 jobs
# In the interest of computational speed we use a minimal basis
# set. For more quantitatively meaningful results, you should use
# a larger basis.
common = Settings()
common.input.ams.Task = "SinglePoint"
common.input.adf.Basis.Type = "SZ"
common.input.adf.Basis.Core = "None"
common.input.adf.Symmetry = "NoSym"

# Specific settings for full system job
full = Settings()
full.input.adf.transferintegrals = True

# Load XYZ file and separate it into 2 fragments
mol = Molecule("BenzeneDimer.xyz")
mol.guess_bonds()
fragments = mol.separate()
if len(fragments) != 2:
    log("ERROR: Molecule {} was split into {} fragments".format(mol.properties.name, len(fragments)))
    import sys

    sys.exit(1)
else:
    mol1, mol2 = fragments

# Alternatively one could simply load fragments from separate xyz files:
# mol1 = Molecule('fragment1.xyz')
# mol2 = Molecule('fragment2.xyz')

job = ADFFragmentJob(name="ADFTI", fragment1=mol1, fragment2=mol2, settings=common, full_settings=full)
results = job.run()

# TI is a dictionary with the whole TransferIntegrals section from adf.rkf
print("== Results ==")
TI = results.get_transfer_integrals()
for key, value in sorted(TI.items()):
    print("{:<28}: {:>12.6f}".format(key, value))
